# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

import json
import logging

from copy import deepcopy

from superdesk import get_resource_service
from eve.utils import ParsedRequest, config
from superdesk.utils import ListCursor
from superdesk.resource import Resource, build_custom_hateoas
from superdesk.services import BaseService
from superdesk.errors import SuperdeskApiError
from superdesk.publish import SUBSCRIBER_TYPES  # NOQA
from flask import current_app as app
from superdesk.metadata.utils import ProductTypes
from superdesk.notification import push_notification


logger = logging.getLogger(__name__)


class SubscribersResource(Resource):
    schema = {
        "name": {"type": "string", "iunique": True, "required": True, "nullable": False, "empty": False},
        "media_type": {"type": "string"},
        "subscriber_type": {"type": "string", "allowed": tuple(SUBSCRIBER_TYPES), "required": True},
        "sequence_num_settings": {
            "type": "dict",
            "schema": {"min": {"type": "integer"}, "max": {"type": "integer"}},
            "required": True,
        },
        "email": {"type": "string", "empty": False, "multiple_emails": True, "required": True},
        "is_active": {"type": "boolean", "default": True},
        "is_targetable": {"type": "boolean", "default": True},
        "critical_errors": {"type": "dict", "valueschema": {"type": "boolean"}},
        "last_closed": {
            "type": "dict",
            "schema": {
                "closed_at": {"type": "datetime"},
                "closed_by": Resource.rel("users", nullable=True),
                "message": {"type": "string"},
            },
        },
        "destinations": {
            "type": "list",
            "schema": {
                "type": "dict",
                "schema": {
                    "name": {"type": "string", "required": True, "empty": False},
                    "format": {"type": "string", "required": True},
                    "preview_endpoint_url": {"type": "string"},
                    "delivery_type": {"type": "string", "required": True},
                    "config": {"type": "dict"},
                },
            },
        },
        "products": {"type": "list", "schema": Resource.rel("products", True)},
        "codes": {"type": "string"},
        "global_filters": {"type": "dict", "valueschema": {"type": "boolean"}},
        "content_api_token": {
            "type": "string",
        },
        "api_products": {"type": "list", "schema": Resource.rel("products", True)},
        "async": {
            "type": "boolean",
            "nullable": True,
        },
        "priority": {
            "type": "boolean",
            "nullable": True,
        },
        "filter_conditions": {
            "type": "list",
        },
        "content_filters": {
            "type": "list",
        },
        "selected_subscribers": {
            "type": "list",
        },
        "init_version": {"type": "integer"},
    }

    item_methods = ["GET", "PATCH", "PUT"]

    privileges = {"POST": "subscribers", "PATCH": "subscribers"}

    mongo_indexes = {
        "name_1": ([("name", 1)], {"unique": True}),
    }


class SubscribersService(BaseService):
    def get(self, req, lookup):
        if req is None:
            req = ParsedRequest()
        if req.args and req.args.get("filter_condition"):
            filter_condition = json.loads(req.args.get("filter_condition"))
            return ListCursor(self._get_subscribers_by_filter_condition(filter_condition))
        return super().get(req=req, lookup=lookup)

    def on_create(self, docs):
        for doc in docs:
            self._validate_seq_num_settings(doc)
            self._validate_products_destinations(doc)

    def on_created(self, docs):
        push_notification("subscriber:create", _id=[doc.get(config.ID_FIELD) for doc in docs])

    def on_update(self, updates, original):
        self._validate_seq_num_settings(updates)
        subscriber = deepcopy(original)
        subscriber.update(updates)
        self._validate_products_destinations(subscriber)

    def on_updated(self, updates, original):
        push_notification("subscriber:update", _id=[original.get(config.ID_FIELD)])

    def on_deleted(self, doc):
        get_resource_service("sequences").delete(
            lookup={"key": "ingest_providers_{_id}".format(_id=doc[config.ID_FIELD])}
        )

    def is_async(self, subscriber_id):
        subscriber = self.find_one(req=None, _id=subscriber_id)
        return subscriber and bool(subscriber.get("async", False))

    def _get_subscribers_by_filter_condition(self, filter_condition):
        """
        Searches all subscribers that has a content filter with the given filter condition

        If filter condition is used in a global filter then it returns all
        subscribers that not disabled the global filter.
        :param filter_condition: Filter condition to test
        :return: List of subscribers
        """
        req = ParsedRequest()
        all_subscribers = list(super().get(req=req, lookup=None))
        selected_products = {}
        selected_subscribers = {}
        selected_content_filters = {}

        filter_condition_service = get_resource_service("filter_conditions")
        content_filter_service = get_resource_service("content_filters")
        product_service = get_resource_service("products")

        existing_products = list(product_service.get(req=req, lookup=None))
        existing_filter_conditions = filter_condition_service.check_similar(filter_condition)
        for fc in existing_filter_conditions:
            existing_content_filters = content_filter_service.get_content_filters_by_filter_condition(fc["_id"])
            for pf in existing_content_filters:
                selected_content_filters[pf["_id"]] = pf

                if pf.get("is_global", False):
                    for s in all_subscribers:
                        gfs = s.get("global_filters", {})
                        if gfs.get(str(pf["_id"]), True):
                            build_custom_hateoas({"self": {"title": "subscribers", "href": "/subscribers/{_id}"}}, s)
                            selected_subscribers[s["_id"]] = s

                for product in existing_products:
                    if (
                        product.get("content_filter")
                        and "filter_id" in product["content_filter"]
                        and product["content_filter"]["filter_id"] == pf["_id"]
                    ):
                        selected_products[product["_id"]] = product

                for s in all_subscribers:
                    all_subscriber_products = list(set(s.get("products") or []) | set(s.get("api_products") or []))
                    for p in all_subscriber_products:
                        if p in selected_products:
                            build_custom_hateoas({"self": {"title": "subscribers", "href": "/subscribers/{_id}"}}, s)
                            selected_subscribers[s["_id"]] = s

        res = {
            "filter_conditions": existing_filter_conditions,
            "content_filters": list(selected_content_filters.values()),
            "products": list(selected_products.values()),
            "selected_subscribers": list(selected_subscribers.values()),
        }
        return [res]

    def _validate_products_destinations(self, subscriber):
        """Validates the subscribers
            1. At least one destination or one api_products is specified.
            2. If direct products are specified then at least one destination is specified.
        :param dict subscriber:
        :return:
        """
        if not subscriber.get("is_active"):
            return

        if not subscriber.get("destinations") and not subscriber.get("api_products"):
            raise SuperdeskApiError.badRequestError(
                payload={"destinations": {"required": 1}, "api_products": {"required": 1}},
                message="At least one destination or one API Product should " "be specified",
            )

        if len(subscriber.get("products") or []) and not subscriber.get("destinations"):
            raise SuperdeskApiError.badRequestError(
                payload={"destinations": {"required": 1}}, message="Destinations not specified."
            )

        if subscriber.get("products"):
            lookup = {config.ID_FIELD: {"$in": subscriber.get("products")}, "product_type": ProductTypes.API.value}
            products = get_resource_service("products").get_product_names(lookup)
            if products:
                raise SuperdeskApiError.badRequestError(
                    payload={"products": 1}, message="Invalid Product Type. " "Products {}.".format(", ".join(products))
                )
        if subscriber.get("api_products"):
            lookup = {
                config.ID_FIELD: {"$in": subscriber.get("api_products")},
                "product_type": ProductTypes.DIRECT.value,
            }
            products = get_resource_service("products").get_product_names(lookup)
            if products:
                raise SuperdeskApiError.badRequestError(
                    payload={"products": 1},
                    message="Invalid Product Type. " "API Products {}.".format(", ".join(products)),
                )

    def get_subscriber_names(self, lookup):
        """Get the subscriber names based on the lookup.
        :param dict lookup: search criteria
        :return list: list of subscriber name
        """
        subscribers = list(self.get(req=None, lookup=lookup))
        return [subscriber["name"] for subscriber in subscribers]

    def _validate_seq_num_settings(self, subscriber):
        """
        Validates the 'sequence_num_settings' property if present in subscriber.

        Below are the validation rules:
            1.  If min value is present then it should be greater than 0
            2.  If min is present and max value isn't available then it's defaulted to MAX_VALUE_OF_PUBLISH_SEQUENCE

        :return: True if validation succeeds otherwise return False.
        """

        if subscriber.get("sequence_num_settings"):
            min = subscriber.get("sequence_num_settings").get("min", 1)
            max = subscriber.get("sequence_num_settings").get("max", app.config["MAX_VALUE_OF_PUBLISH_SEQUENCE"])

            if min <= 0:
                raise SuperdeskApiError.badRequestError(
                    payload={"sequence_num_settings.min": 1},
                    message="Value of Minimum in Sequence Number Settings should " "be greater than 0",
                )

            if min >= max:
                raise SuperdeskApiError.badRequestError(
                    payload={"sequence_num_settings.min": 1},
                    message="Value of Minimum in Sequence Number Settings should " "be less than the value of Maximum",
                )

            del subscriber["sequence_num_settings"]
            subscriber["sequence_num_settings"] = {"min": min, "max": max}

        return True

    def generate_sequence_number(self, subscriber):
        """
        Generates Published Sequence Number for the passed subscriber
        """

        assert subscriber is not None, "Subscriber can't be null"
        min_seq_number = 1
        max_seq_number = app.config["MAX_VALUE_OF_PUBLISH_SEQUENCE"]
        if subscriber.get("sequence_num_settings"):
            min_seq_number = subscriber["sequence_num_settings"]["min"]
            max_seq_number = subscriber["sequence_num_settings"]["max"]

        return get_resource_service("sequences").get_next_sequence_number(
            key_name="subscribers_{_id})".format(_id=subscriber[config.ID_FIELD]),
            max_seq_number=max_seq_number,
            min_seq_number=min_seq_number,
        )



import pytz
import flask
import babel.dates as dates

from apps.auth import get_user
from eve.render import send_response


bp = flask.Blueprint("locales", __name__)


def get_timezones():
    user = get_user()
    lang = user.get("language", flask.current_app.config.get("DEFAULT_LANGUAGE", "en")).replace("-", "_")
    return [
        {
            "id": tz,
            "name": dates.get_timezone_name(tz, locale=lang),
            "location": dates.get_timezone_location(tz, locale=lang),
        }
        for tz in pytz.common_timezones
    ]


@bp.route("/locales/timezones", methods=["GET", "OPTIONS"])
def locales_view():
    if not flask.current_app.auth.authorized([], "locales", "GET"):
        flask.abort(401)
    resp = None
    if flask.request.method == "GET":
        resp = {"timezones": get_timezones()}
    return send_response(None, (resp, None, None, 200))


def init_app(app):
    bp.url_prefix = "/{}".format(app.config["URL_PREFIX"].lstrip("/"))
    app.register_blueprint(bp)



#!/usr/bin/env python3
# This file is part of Superdesk.
#
# Copyright 2019 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

"""This module contains tools to manage content for Superdesk editor"""

import re
import logging
import uuid
import lxml.etree as etree

from textwrap import dedent
from collections.abc import MutableSequence

from flask import current_app as app

from draftjs_exporter.html import HTML
from draftjs_exporter.constants import ENTITY_TYPES, INLINE_STYLES, BLOCK_TYPES
from draftjs_exporter.defaults import STYLE_MAP, BLOCK_MAP
from draftjs_exporter.dom import DOM
from .etree import parse_html, to_string


logger = logging.getLogger(__name__)
DUMMY_RE = re.compile(r"</?dummy_tag>")

ANNOTATION = "ANNOTATION"
MEDIA = "MEDIA"
TABLE = "TABLE"

EDITOR_STATE = "draftjsState"
ENTITY_MAP = "entityMap"
ENTITY_RANGES = "entityRanges"
INLINE_STYLE_RANGES = "inlineStyleRanges"

TAG_STYLE_MAP = {
    "i": INLINE_STYLES.ITALIC,
    "em": INLINE_STYLES.ITALIC,
    "b": INLINE_STYLES.BOLD,
    "strong": INLINE_STYLES.BOLD,
}

TAG_ENTITY_MAP = {
    "a": ENTITY_TYPES.LINK,
}

# FIXME: this is a temporary flag to check that HTML generated by ``generate_fields`` is the
#   same as the one which was already presents if it exists (i.e. the one generated by client).
#   If they differ, a warning will be logged. This is a safety measure to check backend
#   generator before switching off the one from client. More details at SDESK-4911
CHECK_GENERATE_CONSISTENCY = True

# FIXME: those fields are currently hardcoded because they were hardcoded in client too
#   (see https://github.com/superdesk/superdesk-core/pull/1865#issuecomment-632103773).
#   A cleaner way to get field content type should be done (cf. https://dev.sourcefabric.org/browse/SDESK-5316)
TEXT_FIELDS = ["headline", "slugline"]


def get_field_content_state(item, field):
    try:
        return item["fields_meta"][field][EDITOR_STATE][0]
    except (KeyError, AttributeError):
        return None


def get_content_state_fields(item):
    """Return all fields which have a content state"""
    return (field for field in item.get("fields_meta", {}))


def set_field_content_state(item, field, content_state):
    item.setdefault("fields_meta", {}).update({field: {EDITOR_STATE: [content_state]}})


class Entity:
    """Abstraction of a DraftJS entity"""

    def __init__(self, ranges, data):
        self.ranges = ranges
        self.data = data

    @property
    def key(self):
        return str(self.ranges["key"])

    @property
    def type(self):
        return self.data["type"]

    @property
    def is_mutable(self):
        return self.data["mutability"] == "MUTABLE"

    @property
    def content(self):
        return self.data["data"]

    def __str__(self):
        return str(self.data)

    def __repr__(self):
        return "{mutable}{type} entity (key: {key})".format(
            mutable="mutable " if self.is_mutable else "",
            type=self.type,
            key=self.key,
        )


class EntitySequence(MutableSequence):
    def __init__(self, editor, block):
        self._ranges = block.data.setdefault("entityRanges", [])
        self._mapping = editor.content_state.setdefault("entityMap", {})

    def __getitem__(self, key):
        ranges = self._ranges[key]
        data = self._mapping[str(ranges["key"])]
        return Entity(ranges, data)

    def __setitem__(self, key, value):
        if not isinstance(value, Entity):
            raise TypeError("an Entity instance is expected")
        self._ranges[key] = value.ranges
        self._mapping[value.key] = value.data

    def __delitem__(self, key):
        entity_key = str(self._ranges[key]["key"])
        del self._ranges[key]
        del self._mapping[entity_key]

    def __len__(self):
        return len(self._ranges)

    def insert(self, index, value):
        if not isinstance(value, Entity):
            raise TypeError("an Entity instance is expected")
        self._ranges.insert(index, value.ranges)
        self._mapping[value.key] = value.data


class Block:
    """Abstraction of DraftJS block"""

    def __init__(self, editor, data=None, text=None):
        if data is None:
            data = {
                "key": str(uuid.uuid4()),
                "text": text or "",
                "type": "unstyled",
                "depth": 0,
                "inlineStyleRanges": [],
                "entityRanges": [],
                "data": {},
            }
        self.data = data
        self.entities = EntitySequence(editor, self)

    @property
    def depth(self):
        return self.data["depth"]

    @property
    def type(self):
        return self.data["type"]

    @property
    def text(self):
        return self.data.get("text")

    @property
    def key(self):
        return self.data.get("key")

    def __str__(self):
        return self.text

    def __repr__(self):
        return "{type} block with text {text!r} and {len_entities} entities".format(
            type=self.type, text=self.text, len_entities=len(self.entities)
        )


class EmbedBlock(Block):
    def __init__(self, editor, html):
        """
        :param str html: raw HTML to embed
        """
        super().__init__(editor)
        self.data.update(
            {
                "text": " ",
                "type": "atomic",
            }
        )
        entity_data = {
            "type": "EMBED",
            "mutability": "MUTABLE",
            "data": {
                "data": {
                    "html": html,
                }
            },
        }
        entity = Entity(
            ranges={"offset": 0, "length": 1, "key": editor.get_next_entity_key()},
            data=entity_data,
        )
        self.entities.append(entity)


class UnstyledBlock(Block):
    def __init__(self, editor, text=None):
        super().__init__(editor)
        if text:
            self.data.update({"text": text})


class BlockSequence(MutableSequence):
    def __init__(self, editor):
        self.editor = editor
        self._blocks = editor.content_state.setdefault("blocks", {})

    def __getitem__(self, key):
        return Block(self.editor, self._blocks[key])

    def __setitem__(self, key, value):
        if not isinstance(value, Block):
            raise TypeError("a Block instance is expected")
        self._blocks[key] = value.data

    def __delitem__(self, key):
        del self._blocks[key]

    def __len__(self):
        return len(self._blocks)

    def insert(self, index, value):
        if not isinstance(value, Block):
            raise TypeError("a Block instance is expected")
        self._blocks.insert(index, value.data)


class EditorContent:
    """Base class to manage content from editors"""

    def __init__(self):
        raise RuntimeError("This class must not be instantiated directly, use create()")

    @staticmethod
    def create(item, field):
        """Factory for EditorContent"""
        return Editor3Content(item, field)


class DraftJSHTMLExporter:
    def __init__(self, content_editor):
        self.content_editor = content_editor
        # XXX: we need one exporter per DraftJSHTMLExporter because
        #      self.render_media needs to access "entityMap" from local
        #      instance. If MEDIA rendering changes in the future, exporter
        #      should be global for all DraftJSHTMLExporter instances
        self.exporter = HTML(
            {
                "engine": DOM.LXML,
                "style_map": dict(
                    STYLE_MAP,
                    **{
                        INLINE_STYLES.BOLD: "b",
                        INLINE_STYLES.ITALIC: "i",
                        INLINE_STYLES.FALLBACK: self.style_fallback,
                    },
                ),
                "entity_decorators": {
                    ENTITY_TYPES.LINK: self.render_link,
                    ENTITY_TYPES.HORIZONTAL_RULE: lambda props: DOM.create_element("hr"),
                    ENTITY_TYPES.EMBED: self.render_embed,
                    MEDIA: self.render_media,
                    ANNOTATION: self.render_annotation,
                    TABLE: self.render_table,
                },
            }
        )

    @property
    def content_state(self):
        return self.content_editor.content_state

    def render(self):
        blocks = self.content_state["blocks"]
        if not blocks:
            return ""
        if blocks[0].get("text", "-").strip() == "" or blocks[-1].get("text", "-").strip() == "":
            # first and last block may be empty due to client constraints, in this case
            # we must discard them during rendering
            content_state = self.content_state.copy()
            content_state["blocks"] = blocks = blocks[:]
            if blocks[0]["text"].strip() == "" and not blocks[0]["entityRanges"]:
                del blocks[0]
            if blocks and blocks[-1]["text"].strip() == "" and not blocks[-1]["entityRanges"]:
                del blocks[-1]
        else:
            content_state = self.content_state

        try:
            for block in content_state["blocks"]:
                block.setdefault("depth", 0)
                block.setdefault("entityRanges", [])
                block.setdefault("inlineStyleRanges", [])
            html = self.exporter.render(content_state)
        except KeyError as e:
            if e.args == ("text",):
                # "text" may be missing in some case (e.g. comments), and the exporter
                # doesn't support it. To avoid a crash, we render again after
                # filtering out all block elements without "text".
                content_state = self.content_state.copy()
                content_state["blocks"] = [b for b in content_state["blocks"] if "text" in b]
                html = self.exporter.render(content_state)
            else:
                raise e
        # see render_media for details
        return DUMMY_RE.sub("", html)

    def render_annotation(self, props):
        return props["children"]

    def render_media(self, props):
        media_props = props["media"]
        media_type = media_props.get("type", "picture")
        rendition = media_props["renditions"].get("original") or media_props["renditions"]["viewImage"]
        alt_text = media_props.get("alt_text") or ""
        desc = media_props.get("description_text")
        if media_type == "picture":
            embed_type = "Image"
            elt = DOM.create_element("img", {"src": rendition["href"], "alt": alt_text}, props["children"])
        elif media_type == "video":
            embed_type = "Video"
            elt = DOM.create_element(
                "video",
                {"control": "control", "src": rendition["href"], "alt": alt_text, "width": "100%", "height": "100%"},
                props["children"],
            )
        elif media_type == "audio":
            embed_type = "Audio"
            elt = DOM.create_element(
                "audio",
                {"control": "control", "src": rendition["href"], "alt": alt_text, "width": "100%", "height": "100%"},
                props["children"],
            )
        else:
            logger.error("Invalid or not implemented media type: {media_type}".format(media_type=media_type))
            return None

        content = DOM.render(elt)

        if desc:
            content += "<figcaption>{}</figcaption>".format(desc)

        # we need to retrieve the key, there is not straightforward way to do it
        # so we find the key in entityMap with a corresponding value
        embed_key = next(
            k for k, v in self.content_state["entityMap"].items() if v["data"].get("media") == props["media"]
        )

        # <dummy_tag> is needed for the comments, because a root node is necessary
        # it will be removed during rendering.
        embed = DOM.parse_html(
            dedent(
                """\
            <dummy_tag><!-- EMBED START {embed_type} {{id: "editor_{key}"}} -->
            <figure>{content}</figure>
            <!-- EMBED END {embed_type} {{id: "editor_{key}"}} --></dummy_tag>"""
            ).format(embed_type=embed_type, key=embed_key, content=content)
        )

        return embed

    def render_link(self, props):
        if "url" in props:
            attribs = {"href": props["url"]}
        else:
            link_data = props.get("link", {})
            if link_data.get("attachment"):
                attribs = {"data-attachment": link_data["attachment"]}
            elif link_data.get("target"):
                attribs = {"href": link_data["href"], "target": link_data["target"]}
            else:
                attribs = {"href": link_data["href"]}

        return DOM.create_element("a", attribs, props["children"])

    def render_embed(self, props):
        embed_pre_process = app.config.get("EMBED_PRE_PROCESS")
        if embed_pre_process:
            for callback in embed_pre_process:
                callback(props["data"])
        # we use superdesk.etree.parse_html instead of DOM.parse_html as the later modify the content
        # and we use directly the wrapping <div> returned with "content='html'". This works because
        # we use the lxml engine with DraftJSExporter.
        div = parse_html(props["data"]["html"], content="html")
        div.set("class", "embed-block")
        description = props.get("description")
        if description:
            p = DOM.create_element("p", {"class": "embed-block__description"}, description)
            DOM.append_child(div, p)

        return div

    def render_table(self, props):
        num_cols = props["data"]["numCols"]
        num_rows = props["data"]["numRows"]
        with_header = props["data"].get("withHeader", False)
        cells = props["data"]["cells"]
        table = DOM.create_element("table")
        if with_header:
            start_row = 1
            thead = DOM.create_element("thead")
            DOM.append_child(table, thead)
            tr = DOM.create_element("tr")
            DOM.append_child(thead, tr)
            for col_idx in range(num_cols):
                th = DOM.create_element("th")
                DOM.append_child(tr, th)
                try:
                    content_state = cells[0][col_idx]
                except IndexError:
                    continue
                try:
                    content = DOM.parse_html(self.exporter.render(content_state))
                except etree.ParserError:
                    continue
                if content.text or len(content):
                    DOM.append_child(th, content)
        else:
            start_row = 0

        if not with_header or num_rows > 1:
            tbody = DOM.create_element("tbody")
            DOM.append_child(table, tbody)

        for row_idx in range(start_row, num_rows):
            tr = DOM.create_element("tr")
            DOM.append_child(tbody, tr)
            for col_idx in range(num_cols):
                td = DOM.create_element("td")
                DOM.append_child(tr, td)
                try:
                    content_state = cells[row_idx][col_idx]
                except IndexError:
                    continue
                try:
                    content = DOM.parse_html(self.exporter.render(content_state))
                except etree.ParserError:
                    continue
                if content.text or len(content):
                    DOM.append_child(td, content)

        return table

    def style_fallback(self, props):
        type_ = props["inline_style_range"]["style"]
        # we need to use fallback for annotations, has they have suffixes, it's
        # not only "ANNOTATION"
        if type_.startswith("ANNOTATION"):
            attribs = {"annotation-id": type_[11:]}
            return DOM.create_element("span", attribs, props["children"])
        if type_.startswith("COMMENT"):
            # nothing to render for comments
            pass
        else:
            logger.error("No style renderer for {type_!r}".format(type_=type_))


class Editor3Content(EditorContent):
    """Handle content for Superdesk Editor 3, using Draft.js ContentState

    see https://medium.com/@rajaraodv/how-draft-js-represents-rich-text-data-eeabb5f25cf2 for documentation
    on ContentState.
    """

    editor_version = 3
    HTML_EXPORTER = DraftJSHTMLExporter

    def __init__(self, item, field="body_html", is_html=True):
        """
        :param item: item containing Draft.js ContentState
        :param field: field to manage, can be "body_html", "headline", etc.
        :param is_html: boolean to indicate if the field is html or text field
        """
        self.item = item
        self.field = field
        self.is_html = is_html
        self.content_state = get_field_content_state(item, field)
        if not self.content_state:
            self._create_state_from_html(item.get(field))
        self.blocks = BlockSequence(self)
        self.html_exporter = DraftJSHTMLExporter(self)

    def _create_state_from_html(self, value=None):
        self.content_state = {
            "blocks": [],
            "entityMap": {},
        }

        if not value:
            return

        def create_entity(entity_type, data, mutability="MUTABLE"):
            key = len(self.content_state["entityMap"].keys())
            self.content_state["entityMap"][str(key)] = {
                "type": entity_type,
                "mutability": mutability,
                "data": data,
            }
            return key

        def create_atomic_block(entity_type, data):
            block = self.create_block("atomic", text=" ").data
            entity_key = create_entity(entity_type, data)
            block["entityRanges"] = [{"offset": 0, "length": 1, "key": entity_key}]
            block["inlineStyleRanges"] = []
            self.content_state["blocks"].append(block)

        if self.is_html:
            root = parse_html(value, "html")
            for i, elem in enumerate(root):
                try:
                    block_type = next((key for key, val in BLOCK_MAP.items() if val == elem.tag))
                    depth = 0
                except StopIteration:
                    block_type = None
                    depth = 0
                    if elem.tag == "figure":
                        try:
                            m = re.search(r'<!-- EMBED START (?:Image|Video) {id: "(.*)"}', str(root[i - 1]).strip())
                            media = self.item["associations"][m.group(1)]
                            create_atomic_block("MEDIA", {"media": media})
                        except (KeyError, IndexError, AttributeError):
                            create_atomic_block("EMBED", {"data": {"html": to_string(elem, method="html")}})
                        continue
                    elif elem.tag in ("ul", "ol"):
                        pass  # generate block for each li
                    elif elem.text and "<!-- EMBED" in str(elem):
                        continue
                    elif elem.tag == "table":
                        data = {"numCols": 0, "numRows": 0, "withHeader": False, "cells": {}}
                        for row, tr in enumerate(elem.iter("tr")):
                            data["numRows"] += 1
                            data["cells"][row] = {}
                            if row == 0 and len(tr):
                                data["numCols"] = len(tr)
                                data["withHeader"] = tr[0].tag == "th"
                            for col, td in enumerate(tr):
                                data["cells"][row][col] = {
                                    "blocks": [
                                        self.create_block("unstyled", text="".join(td.itertext())).data,
                                    ],
                                    "entityMap": {},
                                }
                        create_atomic_block("TABLE", {"data": data})
                        continue
                    elif elem.tag == "div":
                        html = "".join([to_string(child, method="html") for child in elem])
                        if html.startswith("<html><head>"):
                            html = re.sub(
                                r"<\/head><\/html>",
                                "",
                                html.replace("<html><head>", "", 1),
                            )
                        create_atomic_block("EMBED", {"data": {"html": html}})
                        continue
                    else:
                        logger.warning("ignore block %s", str(elem.tag))
                        continue
                block_text = elem.text or ""
                inline_style_ranges = []
                entity_ranges = []
                for child in elem:
                    child_text = "".join(child.itertext())

                    if child.tag in TAG_STYLE_MAP:
                        inline_style_ranges.append(
                            {
                                "offset": len(block_text),
                                "length": len(child_text),
                                "style": TAG_STYLE_MAP[child.tag],
                            }
                        )

                    if child.tag in TAG_ENTITY_MAP:
                        if not child_text:
                            child_text = " "  # must be non-empty
                        if child.tag == "a":
                            data = {"link": {"href": child.attrib.get("href"), "target": child.attrib.get("target")}}
                        else:
                            data = {}
                        entity_key = create_entity(TAG_ENTITY_MAP[child.tag], data)
                        entity_ranges.append({"offset": len(block_text), "length": len(child_text), "key": entity_key})

                    if child.tag == "li":
                        child_type = (
                            BLOCK_TYPES.UNORDERED_LIST_ITEM if elem.tag == "ul" else BLOCK_TYPES.ORDERED_LIST_ITEM
                        )
                        block = self.create_block(child_type, text=child_text).data
                        block.update(depth=depth)
                        self.content_state["blocks"].append(block)

                    block_text += child_text

                    if child.tail and child.tail.strip():
                        block_text += child.tail

                if elem.tail and elem.tail.strip():
                    block_text += elem.tail

                if block_type:  # no block type for ul/ol
                    block = self.create_block(block_type, text=block_text).data
                    block["inlineStyleRanges"] = inline_style_ranges
                    block["entityRanges"] = entity_ranges
                    self.content_state["blocks"].append(block)
        else:
            for line in value.split("\n"):
                self.content_state["blocks"].append(self.create_block(BLOCK_TYPES.UNSTYLED, text=line).data)

    @property
    def html(self):
        return self.html_exporter.render()

    @property
    def text(self):
        return "\n".join([block.text for block in self.blocks]).strip()

    def get_next_entity_key(self):
        """Return a non existing key for entityMap"""
        return max((int(k) for k in self.content_state["entityMap"].keys()), default=-1) + 1

    def update_item(self):
        self.item[self.field] = self.html if self.is_html else self.text
        set_field_content_state(self.item, self.field, self.content_state)

    def create_block(self, block_type, *args, **kwargs):
        cls_name = "{}Block".format(block_type.capitalize())
        if cls_name in globals():
            cls = globals()[cls_name]
            return cls(self, *args, **kwargs)
        else:
            block = Block(self, *args, **kwargs)
            block.data["type"] = block_type
            return block

    def set_blocks(self, blocks):
        try:
            data = self.blocks[0].data.get("data")  # store internal data from first block
        except IndexError:
            data = {}
        self.content_state["blocks"] = [getattr(block, "data", block) for block in blocks]
        self.blocks = BlockSequence(self)
        if not len(self.blocks):
            self.prepend("Unstyled")
        if not self.blocks[0].data.get("data") and data:
            self.blocks[0].data["data"] = data

    def prepend(self, block_type, *args, **kwargs):
        """Shortcut to prepend a block from its type"""
        block = self.create_block(block_type, *args, **kwargs)
        self.prepend_block(block)
        return block

    def prepend_block(self, block):
        # first block may be empty, this is a client constraint. In this case,
        # we prepend our block after it
        if self.blocks and not self.blocks[0].text.strip():
            index = 1
        else:
            index = 0
        self.blocks.insert(index, block)


def _replace_text(content_state, old, new):
    if not old:
        return
    for block in content_state["blocks"]:
        if block.get("type") == "atomic":
            entity = content_state[ENTITY_MAP][str(block[ENTITY_RANGES][0]["key"])]
            if entity["type"] == "TABLE":
                cells = entity["data"]["data"]["cells"]
                for row in cells.values():
                    for cell in row.values():
                        _replace_text(cell, old, new)
            continue
        if not block.get("text"):
            continue
        end = 0
        while True:
            try:
                index = block["text"].index(old, end)
                end = index + len(old)
                block["text"] = new.join([block["text"][:index], block["text"][end:]])
                for range_field in (ENTITY_RANGES, INLINE_STYLE_RANGES):
                    if block.get(range_field):
                        ranges = []
                        for range_ in block[range_field]:
                            range_end = range_["offset"] + range_["length"]
                            if range_["offset"] > end:  # starting after replaced, move it
                                range_["offset"] += len(new) - len(old)
                                ranges.append(range_)
                            elif range_end <= index:  # starting before replaced text, keep it
                                ranges.append(range_)
                            elif range_["offset"] <= index and range_end >= end:  # contain the text, fix length
                                range_["length"] += len(new) - len(old)
                                ranges.append(range_)
                            else:
                                # remove ranges overlapping with replaced text
                                if range_field == ENTITY_RANGES:
                                    content_state["entityMap"].pop(str(range_["key"]))
                                continue
                        block[range_field] = ranges
            except ValueError:
                break


def replace_text(item, field, old, new, is_html=True):
    """Replace all occurences of old replaced with new.

    It won't replace it in atomic blocks and embeds,
    only text blocks, headings, tables, ul/ol.
    """
    editor = Editor3Content(item, field, is_html)
    _replace_text(editor.content_state, old, new)
    editor.update_item()


def filter_blocks(item, field, filter, is_html=True):
    """Filter content blocks for field.

    It will keep only blocks for which filter returns True.
    """
    editor = Editor3Content(item, field, is_html)
    blocks = []
    for block in editor.blocks:
        if filter(block):
            blocks.append(block)
    editor.set_blocks(blocks)
    editor.update_item()


def generate_fields(item, fields=None):
    """Generate item fields from editor states

    :param item: item containing Draft.js ContentState
    :param fields: fields to generate, None to generate all fields with a content state
    """
    if fields is None:
        fields = get_content_state_fields(item)

    for field in fields:
        old_field = None
        if CHECK_GENERATE_CONSISTENCY:
            old_field = item.get(field)
        editor = Editor3Content(item, field, is_html=field not in TEXT_FIELDS)
        editor.update_item()
        if CHECK_GENERATE_CONSISTENCY:
            if old_field is not None and old_field != item[field]:
                logger.warning(
                    "Generated HTML inconsistency between client and backend, we'll use client one",
                    extra=dict(
                        client=old_field,
                        backend=item[field],
                        field_state=get_field_content_state(item, field),
                    ),
                )
                item[field] = old_field



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014, 2015 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

from superdesk.resource import Resource
from content_api import MONGO_PREFIX

schema = {
    "type": {"type": "string"},
    "subscriber": {"type": "string"},
    "uri": {"type": "string"},
    "items_id": {"type": "string"},
    "version": {"type": "string"},
    "remote_addr": {"type": "string"},
}


class ApiAuditResource(Resource):
    """
    Resource class for the auditing of the API
    """

    schema = schema
    mongo_prefix = MONGO_PREFIX
    internal_resource = True
    mongo_indexes = {"subscriber": [("subscriber", 1)]}



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

import os
from superdesk.tests import TestCase
from superdesk.io.feed_parsers.ana_mpe_newsml import ANANewsMLOneFeedParser
from superdesk.etree import etree


class TestANACase(TestCase):
    def setUp(self):
        dirname = os.path.dirname(os.path.realpath(__file__))
        fixture = os.path.normpath(os.path.join(dirname, "../fixtures", "ana1.xml"))
        provider = {"name": "Test", "source": "ANA"}
        with open(fixture, "rb") as f:
            self.item = ANANewsMLOneFeedParser().parse(etree.fromstring(f.read()), provider)

    def test_headline(self):
        self.assertEqual(self.item["headline"], "Wage employment balance positive in Jan-March")
        self.assertEqual(self.item["subject"][0], {"qcode": "04008004", "name": "economic indicator"})
        self.assertEqual(self.item["dateline"]["text"], "ATHENS, April 4 ANA -")
        self.assertTrue(self.item["body_html"].startswith("<p>Wage employment"))
        self.assertTrue(
            self.item["body_html"].endswith(" job positions.</p>\n<p> </p>"),
        )
        self.assertEqual(self.item["word_count"], "115")



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license


import requests
from datetime import timedelta
from unittest.mock import patch, MagicMock

import superdesk
from superdesk.utc import utcnow
from superdesk.io.registry import register_feeding_service
from superdesk.io.feeding_services.http_service import HTTPFeedingService
from superdesk.tests import TestCase
from superdesk.errors import IngestApiError


TEST_FEEDING_SERVICE_NAME = "test_feeding_service"


def setup_provider(token, hours):
    return {
        "_id": "foo",
        "name": "test http",
        "source": "test http",
        "feeding_service": TEST_FEEDING_SERVICE_NAME,
        "feed_parser": "newsml2",
        "content_expiry": 2880,
        "tokens": {
            "auth_token": token,
            "created": utcnow() - timedelta(hours=hours),
        },
    }


class TestFeedingService(HTTPFeedingService):
    NAME = TEST_FEEDING_SERVICE_NAME
    ERRORS = []

    def _update(self, provider, update):
        pass


register_feeding_service(TestFeedingService)


class ErrorResponseSession(MagicMock):
    def get(self, *args, **kwargs):
        response = requests.Response()
        response.status_code = 401
        return response


class GetTokenTestCase(TestCase):
    def test_get_null_token(self):
        provider = {}
        self.assertEquals("", TestFeedingService()._get_auth_token(provider))

    def test_get_existing_token(self):
        provider = setup_provider("abc", 10)
        self.assertEquals("abc", TestFeedingService()._get_auth_token(provider))

    def test_get_expired_token(self):
        """Expired is better than none.."""
        provider = setup_provider("abc", 24)
        self.assertEquals("", TestFeedingService()._get_auth_token(provider))

    def test_fetch_token(self):
        # TODO: need some rewriting
        # this test is not working anymore
        # try to fill os.environ['REUTERS_USERNAME']
        provider = setup_provider("abc", 24)
        superdesk.get_resource_service("ingest_providers").post([provider])
        self.assertTrue(provider.get("_id"))
        provider["config"] = {}
        provider["config"]["username"] = ""
        provider["config"]["password"] = ""
        # provider['config']['username'] = os.environ.get('REUTERS_USERNAME', '')
        # provider['config']['password'] = os.environ.get('REUTERS_PASSWORD', '')
        # Tests shouldn't depends on some external settings
        # this block should be run always or must be removed %)
        if provider["config"]["username"]:
            token = TestFeedingService()._generate_auth_token(provider, update=True)
            self.assertNotEquals("", token)
            self.assertEquals(token, provider["tokens"]["auth_token"])

            dbprovider = superdesk.get_resource_service("ingest_providers").find_one(name="test http", req=None)
            self.assertEquals(token, dbprovider["tokens"]["auth_token"])

    def test_generate_auth_token_raise_on_error(self):
        provider = setup_provider("abc", 24)
        provider["config"] = {"auth_url": "http://example.com"}
        with patch("requests.Session", new=ErrorResponseSession):
            service = TestFeedingService()
            with self.assertRaises(IngestApiError) as cm:
                service._generate_auth_token(provider)
            self.assertEqual(cm.exception.code, 4007)



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

import logging
import superdesk
from .broadcast import ArchiveBroadcastResource, ArchiveBroadcastService, ARCHIVE_BROADCAST_NAME

logger = logging.getLogger(__name__)


def init_app(app):

    endpoint_name = ARCHIVE_BROADCAST_NAME
    service = ArchiveBroadcastService(endpoint_name, backend=superdesk.get_backend())
    ArchiveBroadcastResource(endpoint_name, app=app, service=service)

    superdesk.privilege(
        name=ARCHIVE_BROADCAST_NAME, label="Broadcast", description="Allows user to create broadcast content."
    )



import superdesk
import traceback
from superdesk import get_resource_service
from .resource import Resource
from .services import BaseService
import logging


logger = logging.getLogger(__name__)


def init_app(app):
    endpoint_name = "sequences"
    service = SequencesService(endpoint_name, backend=superdesk.get_backend())
    SequencesResource(endpoint_name, app=app, service=service)


class SequencesResource(Resource):
    schema = {
        "key": {"type": "string", "required": True, "nullable": False, "empty": False, "iunique": True},
        "sequence_number": {
            "type": "number",
            "default": 1,
        },
    }
    etag_ignore_fields = ["sequence_number", "name"]
    internal_resource = True
    mongo_indexes = {
        "key_1": ([("key", 1)], {"unique": True}),
    }


class SequencesService(BaseService):
    def get_next_sequence_number(self, key_name, max_seq_number=None, min_seq_number=1):
        """
        Generates Sequence Number

        :param key: key to identify the sequence
        :param max_seq_num: default None, maximal possible value, None means no upper limit
        :param min_seq_num: default 1, init value, sequence will start from the NEXT one
        :returns: sequence number
        """
        if not key_name:
            logger.error("Empty sequence key is used: {}".format("\n".join(traceback.format_stack())))
            raise KeyError("Sequence key cannot be empty")

        target_resource = get_resource_service("sequences")
        sequence_number = target_resource.find_and_modify(
            query={"key": key_name}, update={"$inc": {"sequence_number": 1}}, upsert=True, new=True
        ).get("sequence_number")

        if max_seq_number:
            if sequence_number > max_seq_number:
                target_resource.find_and_modify(
                    query={"key": key_name}, update={"$set": {"sequence_number": min_seq_number}}
                )

                sequence_number = min_seq_number

        return sequence_number



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

import superdesk

from copy import deepcopy
from datetime import datetime
from flask import current_app as app
from eve.io.mongo import Validator
from superdesk.metadata.item import ITEM_TYPE
from superdesk.logging import logger
from superdesk.text_utils import get_text
from superdesk import get_resource_service
from _collections_abc import MutableMapping
from superdesk.signals import item_validate
from superdesk.validator import BaseErrorHandler
from flask_babel import lazy_gettext, _


REQUIRED_FIELD = "is a required field"
MAX_LENGTH = "max length is {length}"
STRING_FIELD = "require a string value"
DATE_FIELD = "require a date value"
REQUIRED_ERROR = "{} is a required field"
INVALID_CHAR = "contains invalid characters"
TOO_LONG = "{} is too long"
TOO_SHORT = "{} is too short"
EMPTY_VALUE = "empty values not allowed"

FIELD_LABELS = {
    "abstract": lazy_gettext("Abstract"),
    "alt_text": lazy_gettext("Alt text"),
    "anpa_take_key": lazy_gettext("Take Key"),
    "archive_description": lazy_gettext("Archive description"),
    "attachments": lazy_gettext("Attachments"),
    "authors": lazy_gettext("Authors"),
    "body_footer": lazy_gettext("Body footer"),
    "body_html": lazy_gettext("Body HTML"),
    "byline": lazy_gettext("Byline"),
    "categories": lazy_gettext("Categories"),
    "company_codes": lazy_gettext("Company Codes"),
    "copyrightholder": lazy_gettext("Copyright holder"),
    "copyrightnotice": lazy_gettext("Copyright notice"),
    "dateline": lazy_gettext("Dateline"),
    "description_text": lazy_gettext("Description"),
    "desk": lazy_gettext("Desk"),
    "ednote": lazy_gettext("Ed. Note"),
    "embargo": lazy_gettext("Embargo"),
    "feature_image": lazy_gettext("Feature Image"),
    "feature_media": lazy_gettext("Feature Media"),
    "footer": lazy_gettext("Footer"),
    "footers": lazy_gettext("Footers"),
    "genre": lazy_gettext("Genre"),
    "headline": lazy_gettext("Headline"),
    "ingest_provider": lazy_gettext("Ingest Provider"),
    "keywords": lazy_gettext("Keywords"),
    "language": lazy_gettext("Language"),
    "media": lazy_gettext("Media"),
    "media_description": lazy_gettext("Media Description"),
    "place": lazy_gettext("Place"),
    "priority": lazy_gettext("Priority"),
    "publish_schedule": lazy_gettext("Scheduled Time"),
    "relatedItems": lazy_gettext("Related Items"),
    "sign_off": lazy_gettext("Sign Off"),
    "slugline": lazy_gettext("Slugline"),
    "sms": lazy_gettext("SMS"),
    "sms_message": lazy_gettext("SMS Message"),
    "source": lazy_gettext("Source"),
    "stage": lazy_gettext("Stage"),
    "subject": lazy_gettext("Subject"),
    "type": lazy_gettext("Type"),
    "urgency": lazy_gettext("Urgency"),
    "usageterms": lazy_gettext("Usage Terms"),
}

ERROR_MESSAGES = {
    REQUIRED_FIELD: lazy_gettext("is a required field"),
    MAX_LENGTH: lazy_gettext("max length is {length}"),
    STRING_FIELD: lazy_gettext("require a string value"),
    DATE_FIELD: lazy_gettext("require a date value"),
    REQUIRED_ERROR: lazy_gettext("{} is a required field"),
    INVALID_CHAR: lazy_gettext("contains invalid characters"),
    TOO_LONG: lazy_gettext("{} is too long"),
    TOO_SHORT: lazy_gettext("{} is too short"),
    EMPTY_VALUE: lazy_gettext("empty values not allowed"),
}


def check_json(doc, field, value):
    if isinstance(doc, dict):
        if field in doc and value["required"] and doc[field]:
            return True
        for key in doc:
            if check_json(doc[key], field, value):
                return True
        return False
    elif isinstance(doc, list):
        for item in doc:
            if check_json(item, field, value):
                return True
        return False
    else:
        return False


def get_validator_schema(schema):
    """Get schema for given data that will work with validator.

    - if field is required without minlength set make sure it's not empty
    - if there are keys with None value - remove them

    :param schema
    """
    validator_schema = {key: val for key, val in schema.items() if val is not None}

    if validator_schema.get("required") and not validator_schema.get("minlength"):
        validator_schema.setdefault("empty", False)

    return validator_schema


class SchemaValidator(Validator):
    def __init__(self, *args, **kwargs):
        kwargs["error_handler"] = BaseErrorHandler
        super(SchemaValidator, self).__init__(*args, **kwargs)

    types_mapping = {k: v for k, v in Validator.types_mapping.items() if k not in ("date",)}

    def _validate_type_media(self, value):
        return True

    def _validate_type_related_content(self, value):
        return True

    def _validate_type_embed(self, value):
        return True

    def _validate_type_custom(self, value):
        return True

    def _validate_type_picture(self, value):
        return True

    def _validate_type_any(self, value):
        """Allow type any, ex: for CV of type 'custom'."""
        return True

    def _validate_type_text(self, value):
        """Validate type text in schema."""
        if value and isinstance(value, str):
            return True

    def _validate_type_date(self, value):
        if isinstance(value, datetime):
            return True
        try:
            datetime.strptime(value or "", "%Y-%m-%dT%H:%M:%S+%f")
            return True
        except ValueError:
            pass

    def _validate_mandatory_in_list(self, mandatory, field, value):
        """
        {'type': 'dict'}
        """
        subject_schemas = {}

        if field == "subject":
            cvs = get_resource_service("vocabularies").get_from_mongo(req=None, lookup={"schema_field": field})
            for cv in cvs:
                subject_schemas.update({field: cv["_id"]})

        for key in mandatory:
            for key_field in mandatory[key]:
                if mandatory[key][key_field]["required"]:
                    if (
                        value
                        and len(value) > 0
                        and any(v["scheme"] == key_field or key_field in subject_schemas for v in value)
                    ):
                        if not check_json(value, key, mandatory[key][key_field]):
                            self._error(key_field, REQUIRED_FIELD)
                    else:
                        self._error(key_field, REQUIRED_FIELD)

    def _validate_mandatory_in_dictionary(self, mandatory, field, value):
        """
        {'type': 'list'}
        """
        for key in mandatory:
            if not value.get(key):
                self._error(key, REQUIRED_FIELD)

    def _validate_empty(self, empty, field, value):
        """
        {'type': 'boolean'}
        """
        super()._validate_empty(empty, field, value)
        if field == "subject":
            # for subject, we have to ignore all data with scheme
            # as they are used for custom values except "subject_custom" scheme as it's the scheme for subject cv
            # so it must be present
            subject_schemas = {None, "", "subject_custom"}

            # plus any cv with schema_field subject
            cvs = get_resource_service("vocabularies").get_from_mongo(
                req=None, lookup={"schema_field": field}, projection={"_id": 1}
            )
            for cv in cvs:
                subject_schemas.add(cv["_id"])

            filtered = [v for v in value if v.get("scheme") in subject_schemas]

            if not filtered:
                self._error(field, REQUIRED_FIELD)

        elif isinstance(value, list) and not value:
            self._error(field, REQUIRED_FIELD)
        elif isinstance(value, str) and value == "<p></p>":  # default value for editor3
            self._error(field, REQUIRED_FIELD)

    def _validate_enabled(self, *args):
        """
        {'type': 'boolean'}
        """
        pass

    def _validate_readonly(self, *args):
        """
        {'type': 'boolean'}
        """
        pass

    def _validate_place(self, *args):
        """
        {'type': 'list'}
        """
        pass

    def _validate_genre(self, *args):
        """
        {'type': 'dict'}
        """
        pass

    def _validate_anpa_category(self, *args):
        """
        {'type': 'list'}
        """
        pass

    def _validate_subject(self, *args):
        """
        {'type': 'list'}
        """
        pass

    def _validate_company_codes(self, *args):
        """
        {'type': 'list'}
        """
        pass

    def _validate_validate_characters(self, validate, field, value):
        """
        {'type': 'boolean'}
        """
        disallowed_characters = app.config.get("DISALLOWED_CHARACTERS")

        if validate and disallowed_characters and value:
            invalid_chars = [char for char in disallowed_characters if char in value]
            if invalid_chars:
                return self._error(field, INVALID_CHAR)

    def _validate_media_metadata(self, validate, associations_field, associations):
        """
        {'type': 'boolean'}
        """
        if not validate:
            return
        media_metadata_schema = app.config.get("VALIDATOR_MEDIA_METADATA")
        if not media_metadata_schema:
            return
        for assoc_name, assoc_data in associations.items():
            if assoc_data is None or assoc_data.get("type") == "text":
                continue
            for field, schema in media_metadata_schema.items():
                if schema.get("required", False) and not assoc_data.get(field):
                    self._error("media's " + field, REQUIRED_FIELD)
                try:
                    max_length = int(schema["maxlength"])
                except KeyError:
                    pass
                except ValueError:
                    logger.error("Invalid max length value for media field {field}".format(field=field))
                else:
                    if assoc_data.get(field) is not None and len(assoc_data[field]) > max_length:
                        self._error("media's " + field, MAX_LENGTH.format(length=max_length))


class ValidateResource(superdesk.Resource):
    schema = {
        "act": {"type": "string", "required": True},
        "type": {"type": "string", "required": True},
        "embedded": {"type": "boolean", "required": False},
        "validate": {
            "type": "dict",
            "required": True,
        },
        "errors": {"type": "list"},
    }

    resource_methods = ["POST"]
    item_methods = []


class ValidateService(superdesk.Service):
    def create(self, docs, fields=False, **kwargs):
        for doc in docs:
            doc["errors"] = self.validate(doc, fields=fields, **kwargs)
        if fields:
            return [doc["errors"] for doc in docs]
        return [i for i in range(len(docs))]

    def validate(self, doc, fields=False, **kwargs):
        test_doc = deepcopy(doc)
        return self._validate(test_doc, fields=fields, **kwargs)

    def _get_profile_schema(self, schema, doc):
        doc["validate"].setdefault("extra", {})  # make sure extra is there so it will validate its fields
        extra_field_types = {"text": "string", "embed": "dict", "date": "date", "urls": "list", "custom": "any"}
        extra_fields = superdesk.get_resource_service("vocabularies").get_extra_fields()
        schema["extra"] = {"type": "dict", "schema": {}}
        for extra_field in extra_fields:
            if schema.get(extra_field["_id"]) and extra_field.get("field_type", None) in extra_field_types:
                rules = schema.pop(extra_field["_id"])
                rules["type"] = extra_field_types.get(extra_field["field_type"], "string")
                schema["extra"]["schema"].update({extra_field["_id"]: get_validator_schema(rules)})
                self._populate_extra(doc["validate"], extra_field["_id"])
        try:
            # avoid errors when cv is removed and value is still there
            schema["subject"]["schema"]["schema"]["scheme"].pop("allowed", None)
        except KeyError:
            pass
        return [{"schema": schema}]

    def _get_validators(self, doc):
        """Get validators.

        In case there is profile defined for item with respective content type it will
        use its schema for validations, otherwise it will fall back to action/item_type filter.
        """
        profile = doc["validate"].get("profile")
        item_type = doc["validate"].get("type") or doc.get("type")
        if (profile or item_type) and (app.config["AUTO_PUBLISH_CONTENT_PROFILE"] or doc["act"] != "auto_publish"):
            content_type = superdesk.get_resource_service("content_types").find_one(req=None, _id=profile)
            if not content_type and item_type:
                content_type = superdesk.get_resource_service("content_types").find_one(req=None, item_type=item_type)
            if content_type:
                return self._get_profile_schema(content_type.get("schema", {}), doc)
        lookup = {"act": doc["act"], "type": doc[ITEM_TYPE]}
        if doc.get("embedded"):
            lookup["embedded"] = doc["embedded"]
        else:
            lookup["$or"] = [{"embedded": {"$exists": False}}, {"embedded": False}]
        custom_schema = app.config.get("SCHEMA", {}).get(doc[ITEM_TYPE])
        if custom_schema:  # handle custom schema like profile schema
            return self._get_profile_schema(custom_schema, doc)
        return superdesk.get_resource_service("validators").get(req=None, lookup=lookup)

    def _populate_extra(self, doc, schema):
        """Populates the extra field in the document with fields stored in subject. Used
        for user defined vocabularies.

        :param doc: Article to be validated
        :param schema: Vocabulary identifier
        """
        for subject in doc.get("subject", []):
            if subject.get("scheme", "") == schema:
                doc["extra"][schema] = subject.get("qcode", "")

    def _sanitize_fields(self, doc, validator):
        """If maxlength or minlength is specified in the validator then remove any markups from that field

        :param doc: Article to be validated
        :param validator: Validation rule
        :return: updated article
        """
        fields_to_check = ["minlength", "maxlength"]
        item_schema = validator.get("schema", {})
        extra_schema = item_schema.get("extra", {}).get("schema", {})
        schemes_docs = [(item_schema, doc), (extra_schema, doc.get("extra", {}))]
        for schema, content in schemes_docs:
            for field in schema:
                if (
                    content.get(field)
                    and schema.get(field)
                    and type(content[field]) is str
                    and any(k in schema[field] for k in fields_to_check)
                ):
                    try:
                        content[field] = get_text(content[field])
                    except (ValueError, TypeError):
                        # fails for json fields like subject, genre
                        pass

    def _get_media_fields(
        self,
        field_schema,
        field_associations,
        doc,
    ):
        """Returns the field names in associations. For multiple valued media fields
        the field asssociations name is different from the field schema name.
        :param field_schema: Field schema name
        :param field_associations: Field associations name
        :param doc: Article to be validated
        """

        fields = []
        if field_associations in doc.get("associations", {}):
            return [field_associations]
        media_multivalue_field = field_schema + "--"
        for media_field in doc.get("associations", {}):
            if media_field.startswith(media_multivalue_field) and doc.get("associations", {}).get(media_field):
                fields.append(media_field)
        return fields

    def _process_media(self, doc, validation_schema):
        """If media field(feature media or custom media field) is required it should be present for
        validation on doc not only on associations
        If media_description is required it should be present for validation on doc not only on
        associations->featuremedia (or custom media field)
        :param doc: Article to be validated
        :param schema: Schema to validate the article
        """

        for field_schema in validation_schema:
            field_associations = field_schema if field_schema != "feature_media" else "featuremedia"
            media_fields = self._get_media_fields(field_schema, field_associations, doc)
            for media_field in media_fields:
                if media_field and isinstance(doc["associations"][media_field], MutableMapping):
                    doc[field_schema] = doc["associations"][media_field]
                    if media_field != "featuremedia":
                        del doc["associations"][media_field]
                    elif not doc.get("feature_media", None) is None and "description_text" in doc["feature_media"]:
                        doc["media_description"] = doc["feature_media"]["description_text"]

    def _process_sms(self, doc, schema):
        """Apply the SMS validation to the sms_message value if the document is flagged for SMS
        :param doc:
        :param schema:
        :return:
        """

        if doc.get("flags", {}).get("marked_for_sms", False):
            doc["sms"] = doc.get("sms_message", "")
        else:
            # remove it from the valiadation it is not required
            schema.pop("sms", None) if schema.get("sms") else schema.pop("sms_message", None)

    def _process_media_metadata(self, doc, schema):
        """Request media validation if associations is present

        :param doc:
        :param schema:
        :return:
        """
        if doc.get("associations"):
            schema.setdefault("associations", {})["media_metadata"] = app.config.get(
                "VALIDATE_MEDIA_METADATA_ON_PUBLISH", True
            )

    def _get_validator_schema(self, validator):
        """Get schema for given validator.

        And make sure there is no `None` value which would raise an exception.
        """
        return {field: get_validator_schema(schema) for field, schema in validator["schema"].items() if schema}

    def _get_vocabulary_display_name(self, vocabulary_id):
        if vocabulary_id == "anpa_category":
            vocabulary = get_resource_service("vocabularies").find_one(req=None, _id="categories")
        elif vocabulary_id == "subject":
            cv = get_resource_service("vocabularies").find_one(req=None, schema_field=vocabulary_id)
            id = cv["_id"] if cv else vocabulary_id
            vocabulary = get_resource_service("vocabularies").find_one(req=None, _id=id)
        else:
            vocabulary = get_resource_service("vocabularies").find_one(req=None, _id=vocabulary_id)
        if vocabulary and "display_name" in vocabulary:
            return vocabulary["display_name"]
        return vocabulary_id

    def _set_default_subject_scheme(self, item):
        if item.get("subject"):
            for subject in item["subject"]:
                subject.setdefault("scheme", None)

    def get_error_field_name(self, error_field):
        if FIELD_LABELS.get(error_field):
            return FIELD_LABELS.get(error_field)
        return error_field

    def _validate(self, doc, fields=False, **kwargs):
        item = deepcopy(doc["validate"])  # make a copy for signal before validation processing
        use_headline = kwargs and "headline" in kwargs
        validators = self._get_validators(doc)
        for validator in validators:
            validation_schema = self._get_validator_schema(validator)
            self._sanitize_fields(doc["validate"], validator)
            self._set_default_subject_scheme(doc["validate"])
            self._process_media(doc["validate"], validation_schema)
            self._process_sms(doc["validate"], validation_schema)
            self._process_media_metadata(doc["validate"], validation_schema)
            v = SchemaValidator()
            v.allow_unknown = True
            try:
                v.validate(doc["validate"], validation_schema)
            except TypeError as ex:
                logger.exception('Invalid validator schema value "%s" for ' % str(ex))
            error_list = v.errors
            response = []
            for e in error_list:
                messages = []
                # Ignore dateline if item is corrected because it can't be changed after the item is published
                if doc.get("act", None) == "correct" and e == "dateline":
                    continue
                elif (
                    doc.get("act", None) == "kill"
                    and doc["validate"].get("profile", None)
                    and e in ("headline", "abstract", "body_html")
                ):
                    continue
                elif e == "extra":
                    for field in error_list[e]:
                        display_name = self._get_vocabulary_display_name(field)
                        if "required" in error_list[e][field]:
                            messages.append(ERROR_MESSAGES[REQUIRED_ERROR].format(display_name))
                        else:
                            error_field = self.get_error_field_name(display_name)
                            messages.append("{} {}".format(error_field, error_list[e][field]))
                elif "required field" in error_list[e] or type(error_list[e]) is dict or type(error_list[e]) is list:
                    display_name = self._get_vocabulary_display_name(e)
                    error_field = self.get_error_field_name(display_name)
                    messages.append(ERROR_MESSAGES[REQUIRED_ERROR].format(error_field.upper()))
                elif "min length is 1" == error_list[e] or "null value not allowed" in error_list[e]:
                    messages.append(ERROR_MESSAGES[REQUIRED_ERROR].format(e.upper()))
                elif "min length is" in error_list[e]:
                    error_field = self.get_error_field_name(e)
                    messages.append(ERROR_MESSAGES[TOO_SHORT].format(error_field.upper()))
                elif "max length is" in error_list[e]:
                    error_field = self.get_error_field_name(e)
                    messages.append(ERROR_MESSAGES[TOO_LONG].format(error_field.upper()))
                else:
                    error_field = self.get_error_field_name(e)
                    messages.append(
                        "{} {}".format(
                            error_field.upper(),
                            ERROR_MESSAGES[error_list[e]] if ERROR_MESSAGES.get(error_list[e]) else error_list[e],
                        )
                    )

                for message in messages:
                    if use_headline:
                        headline = "{}: {}".format(doc["validate"].get("headline", doc["validate"].get("_id")), message)
                        response.append(headline)
                    else:
                        response.append(message)

            # let custom code do additional validation
            item_validate.send(self, item=item, response=response, error_fields=v.errors)

            if fields:
                return response, v.errors
            return response
        else:
            logger.warn("validator was not found for {}".format(doc["act"]))
            if fields:
                return [], {}
            return []



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

import superdesk

from .app_prepopulate import PrepopulateService, PrepopulateResource
from .app_populate import AppPopulateCommand  # NOQA
from .app_initialize import AppInitializeWithDataCommand  # NOQA
from .app_scaffold_data import AppScaffoldDataCommand  # NOQA


def init_app(app):
    if superdesk.app.config.get("SUPERDESK_TESTING", False):
        endpoint_name = "prepopulate"
        service = PrepopulateService(endpoint_name, backend=superdesk.get_backend())
        PrepopulateResource(endpoint_name, app=app, service=service)

        superdesk.intrinsic_privilege(resource_name=endpoint_name, method=["POST"])



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license


import os

from superdesk.io.feed_parsers.dpa_iptc7901 import DPAIPTC7901FeedParser
from superdesk.tests import TestCase


def fixture(filename):
    dirname = os.path.dirname(os.path.realpath(__file__))
    return os.path.normpath(os.path.join(dirname, "../fixtures", filename))


class DPAIptcTestCase(TestCase):

    parser = DPAIPTC7901FeedParser()

    def open(self, filename):
        provider = {"name": "Test"}
        return self.parser.parse(fixture(filename), provider)

    def test_open_iptc7901_file(self):
        with self.app.app_context():
            item = self.open("IPTC7901.txt")
            self.assertEqual("text", item["type"])
            self.assertEqual("062", item["ingest_provider_sequence"])
            self.assertEqual("i", item["anpa_category"][0]["qcode"])
            self.assertEqual(211, item["word_count"])
            self.assertEqual("Germany Social Democrats: Coalition talks with Merkel could fail", item["headline"])
            self.assertRegex(item["body_html"], "^<p></p><p>Negotiations")
            self.assertEqual("Germany-politics", item["slugline"])
            self.assertEqual(4, item["priority"])
            self.assertEqual([{"qcode": "i"}], item["anpa_category"])
            self.assertTrue(item["ednote"].find("## Editorial contacts"))
            self.assertEqual(item["dateline"]["source"], "dpa")
            self.assertEqual(item["dateline"]["located"]["city"], "Berlin")

    def test_open_dpa_copyright(self):
        with self.app.app_context():
            item = self.open("dpa_copyright.txt")
            self.assertEqual("text", item["type"])
            self.assertEqual("rs", item["anpa_category"][0]["qcode"])
            self.assertEqual("Impressum", item["headline"])
            self.assertEqual("Impressum", item["slugline"])
            self.assertEqual("(Achtung)", item["anpa_take_key"])

    def test_four_line_header(self):
        with self.app.app_context():
            item = self.open("dpa_four_line.txt")
            self.assertEqual("Switzerland joins list of countries resisting UN migration pact", item["headline"])
            self.assertEqual("Switzerland/migration/UN", item["slugline"])
            self.assertEqual("REFILE 1ST LEAD", item["anpa_take_key"])

    def test_two_line_header(self):
        with self.app.app_context():
            item = self.open("dpa_two_line.txt")
            self.assertEqual("Peace talks on Yemen to take place in Sweden next month, Mattis says", item["headline"])
            self.assertEqual("Yemen/conflict", item["slugline"])
            self.assertEqual("EXTRA", item["anpa_take_key"])

    def test_two_by_line_header(self):
        with self.app.app_context():
            item = self.open("dpa_two_by_line.txt")
            self.assertEqual("Chiefs-Rams 2.0? Would be a fun Super Bowl, but don't count on it", item["headline"])
            self.assertEqual("American Football/NFL", item["slugline"])



import copy
import bson

from unittest import mock
from datetime import timedelta
from superdesk.tests import TestCase
from superdesk.utc import utcnow
from apps.content_types import apply_schema

import apps.content_types.content_types as content_types


class MockService:
    def find_one(self, req, **lookup):
        return {"schema": {"headline": {}}}


class MockVocabulariesService:
    def get_forbiden_custom_vocabularies(self):
        return [
            {
                "_id": "test",
                "field_type": None,
                "selection_type": "do not show",
                "display_name": "test",
                "helper_text": "test",
                "unique_field": "qcode",
            }
        ]


class ContentTypesTestCase(TestCase):
    def test_apply_schema_default(self):
        item = {"guid": "guid", "headline": "foo"}
        self.assertEqual(item, apply_schema(item))

    @mock.patch("apps.content_types.content_types.get_resource_service", return_value=MockService())
    def test_apply_schema_profile(self, mock):
        item = {"headline": "foo", "slugline": "bar", "guid": "1", "profile": "test"}
        self.assertEqual({"headline": "foo", "guid": "1", "profile": "test"}, apply_schema(item))

    @mock.patch.object(content_types, "get_fields_map_and_names", lambda: ({}, {}))
    def test_minlength(self):
        """Check that minlength is not modified when it is set

        check is done with required set
        """
        original = {
            "schema": {
                "body_html": {"required": True, "enabled": True},
            }
        }
        updates = copy.deepcopy(original)
        updates["schema"]["body_html"]["minlength"] = "99"
        content_types.ContentTypesService().on_update(updates, original)
        self.assertEqual(updates["schema"]["body_html"]["minlength"], "99")

    def test_get_output_name(self):
        _id = bson.ObjectId()
        service = content_types.ContentTypesService()
        with mock.patch.object(service, "find_one", return_value={"label": "Test Label 123 *#$"}):
            self.assertEqual("TestLabel123", service.get_output_name(_id))

    @mock.patch("apps.content_types.content_types.get_resource_service", return_value=MockVocabulariesService())
    def test_clean_doc(self, mock):
        profile = {
            "schema": {"test": {"type": "list", "required": False, "default": []}},
            "editor": {"test": {"enabled": False, "field_name": "test"}},
        }

        content_types.clean_doc(profile)
        self.assertEqual({"schema": {}, "editor": {}}, profile)

    # Do not remove or modify the following line, the test would always pass without
    # the dict with "hashtags" returned by "get_fields_map_and_names"
    @mock.patch.object(content_types, "get_fields_map_and_names", lambda: ({"hashtags": "hashtags"}, {}))
    def test_subject(self):
        """Check that subject is not set if it's not present in editor (SDESK-3745)

        If we had custom vocabularies in schema, subject was added to "schema" even if not
        present if "editor", resulting in validation error.
        """
        original = {
            "_id": "5b1a4774b10de731297716ad",
            "editor": {
                "hashtags": {"order": 5, "enabled": True, "field_name": "Hashtags"},
                "subject": {"order": 1, "sdWidth": "full", "required": True, "enabled": True},
            },
            "schema": {
                "subject": {
                    "mandatory_in_list": {"scheme": {}},
                    "schema": {},
                    "type": "list",
                    "required": True,
                    "default": [],
                    "nullable": False,
                },
                "hashtags": {"type": "list", "required": False, "default": []},
            },
        }

        updates = copy.deepcopy(original)
        # the updates are only removing the "subject" field
        updates["schema"]["subject"] = updates["editor"]["subject"] = None
        content_types.ContentTypesService().on_update(updates, original)
        self.assertFalse(updates["schema"]["subject"]["required"])

    def test_prepare_for_edit_updated_now(self):
        doc = {
            "editor": {},
            "schema": {},
            "_updated": utcnow() - timedelta(days=5),
        }

        content_types.prepare_for_edit_content_type(doc)
        self.assertGreaterEqual(doc["_updated"], utcnow() - timedelta(seconds=2))



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014, 2015 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

import superdesk
from .resource import ApiAuditResource
from .service import ApiAuditService


def init_app(app):
    """
    Initialise the auditing of the API
    :param app:
    :return:
    """
    endpoint_name = "api_audit"

    service = ApiAuditService(endpoint_name, backend=superdesk.get_backend())
    ApiAuditResource(endpoint_name, app=app, service=service)



#!/usr/bin/env python
# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014, 2015 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

import os
import eve
import flask
import jinja2
import importlib
import superdesk
import logging

from flask_mail import Mail
from eve.auth import TokenAuth
from eve.io.mongo.mongo import _create_index as create_index
from eve.render import send_response
from flask_babel import Babel
from flask import g
from babel import parse_locale
from pymongo.errors import DuplicateKeyError

from superdesk.celery_app import init_celery
from superdesk.datalayer import SuperdeskDataLayer  # noqa
from superdesk.errors import SuperdeskError, SuperdeskApiError
from superdesk.factory.sentry import SuperdeskSentry
from superdesk.logging import configure_logging
from superdesk.storage import ProxyMediaStorage
from superdesk.validator import SuperdeskValidator
from superdesk.json_utils import SuperdeskJSONEncoder

SUPERDESK_PATH = os.path.abspath(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))

logger = logging.getLogger(__name__)


def set_error_handlers(app):
    """Set error handlers for the given application object.

    Each error handler receives a :py:class:`superdesk.errors.SuperdeskError`
    instance as a parameter and returns a tuple containing an error message
    that is sent to the client and the HTTP status code.

    :param app: an instance of `Eve <http://python-eve.org/>`_ application
    """

    @app.errorhandler(SuperdeskError)
    def client_error_handler(error):
        error_dict = error.to_dict()
        error_dict.update(internal_error=error.status_code)
        status_code = error.status_code or 422
        return send_response(None, (error_dict, None, None, status_code))

    @app.errorhandler(403)
    def server_forbidden_handler(error):
        return send_response(None, ({"code": 403, "error": error.response}, None, None, 403))

    @app.errorhandler(AssertionError)
    def assert_error_handler(error):
        print("error", error)
        return send_response(None, ({"code": 400, "error": str(error) if str(error) else "assert"}, None, None, 400))

    @app.errorhandler(500)
    def server_error_handler(error):
        """Log server errors."""
        return_error = SuperdeskApiError.internalError(error)
        return client_error_handler(return_error)


class SuperdeskEve(eve.Eve):
    def __getattr__(self, name):
        """Workaround for https://github.com/pyeve/eve/issues/1087"""
        if name in {"im_self", "im_func"}:
            raise AttributeError("type object '%s' has no attribute '%s'" % (self.__class__.__name__, name))
        return super(SuperdeskEve, self).__getattr__(name)

    def init_indexes(self, ignore_duplicate_keys=False):
        for resource, resource_config in self.config["DOMAIN"].items():
            mongo_indexes = resource_config.get("mongo_indexes__init")
            if not mongo_indexes:
                continue

            # Borrowed https://github.com/pyeve/eve/blob/22ea4bfebc8b633251cd06837893ff699bd07a00/eve/flaskapp.py#L915
            for name, value in mongo_indexes.items():
                if isinstance(value, tuple):
                    list_of_keys, index_options = value
                else:
                    list_of_keys = value
                    index_options = {}

                # index creation in background
                index_options.setdefault("background", True)

                try:
                    create_index(self, resource, name, list_of_keys, index_options)
                except DuplicateKeyError as err:
                    # Duplicate key for unique indexes are generally caused by invalid documents in the collection
                    # such as multiple documents not having a value for the attribute used for the index
                    # Log the error so it can be diagnosed and fixed
                    logger.exception(err)

                    if not ignore_duplicate_keys:
                        raise


def get_app(config=None, media_storage=None, config_object=None, init_elastic=None):
    """App factory.

    :param config: configuration that can override config from ``default_settings.py``
    :param media_storage: media storage class to use
    :param config_object: config object to load (can be module name, module or an object)
    :param init_elastic: obsolete config - kept there for BC
    :return: a new SuperdeskEve app instance
    """

    abs_path = SUPERDESK_PATH
    app_config = flask.Config(abs_path)
    app_config.from_object("superdesk.default_settings")
    app_config.setdefault("APP_ABSPATH", abs_path)
    app_config.setdefault("DOMAIN", {})
    app_config.setdefault("SOURCES", {})

    if config_object:
        app_config.from_object(config_object)

    try:
        app_config.update(config or {})
    except TypeError:
        app_config.from_object(config)

    if not media_storage:
        media_storage = ProxyMediaStorage

    app = SuperdeskEve(
        data=SuperdeskDataLayer,
        auth=TokenAuth,
        media=media_storage,
        settings=app_config,
        json_encoder=SuperdeskJSONEncoder,
        validator=SuperdeskValidator,
        template_folder=os.path.join(abs_path, "templates"),
    )

    app.notification_client = None

    app.jinja_options = {"autoescape": False}
    app.json_encoder = SuperdeskJSONEncoder  # seems like eve param doesn't set it on flask

    # init client_config with default config
    app.client_config = {
        "content_expiry_minutes": app.config.get("CONTENT_EXPIRY_MINUTES", 0),
        "ingest_expiry_minutes": app.config.get("INGEST_EXPIRY_MINUTES", 0),
    }

    superdesk.app = app

    custom_loader = jinja2.ChoiceLoader(
        [
            jinja2.FileSystemLoader("templates"),
            jinja2.FileSystemLoader(os.path.join(SUPERDESK_PATH, "templates")),
        ]
    )

    app.jinja_loader = custom_loader
    app.mail = Mail(app)
    app.sentry = SuperdeskSentry(app)

    # setup babel
    app.config.setdefault("BABEL_TRANSLATION_DIRECTORIES", os.path.join(SUPERDESK_PATH, "translations"))
    app.babel_tzinfo = None
    app.babel_locale = None
    app.babel_translations = None
    babel = Babel(app, configure_jinja=False)

    @babel.localeselector
    def get_locale():
        user = getattr(g, "user", {})
        user_language = user.get("language", app.config.get("DEFAULT_LANGUAGE", "en"))
        try:
            # Attempt to load the local using Babel.parse_local
            parse_locale(user_language.replace("-", "_"))
        except ValueError:
            # If Babel fails to recognise the locale, then use the default language
            user_language = app.config.get("DEFAULT_LANGUAGE", "en")

        return user_language.replace("-", "_")

    set_error_handlers(app)

    @app.after_request
    def after_request(response):
        # fixing previous media prefixes if defined
        if app.config["MEDIA_PREFIXES_TO_FIX"] and app.config["MEDIA_PREFIX"]:
            current_prefix = app.config["MEDIA_PREFIX"].rstrip("/").encode()
            for prefix in app.config["MEDIA_PREFIXES_TO_FIX"]:
                response.data = response.data.replace(prefix.rstrip("/").encode(), current_prefix)
        return response

    init_celery(app)
    installed = set()

    def install_app(module_name):
        if module_name in installed:
            return
        installed.add(module_name)
        app_module = importlib.import_module(module_name)
        if hasattr(app_module, "init_app"):
            app_module.init_app(app)

    for module_name in app.config.get("CORE_APPS", []):
        install_app(module_name)

    for module_name in app.config.get("INSTALLED_APPS", []):
        install_app(module_name)

    for resource in superdesk.DOMAIN:
        app.register_resource(resource, superdesk.DOMAIN[resource])

    for name, jinja_filter in superdesk.JINJA_FILTERS.items():
        app.jinja_env.filters[name] = jinja_filter

    configure_logging(app.config["LOG_CONFIG_FILE"])

    return app



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license


import logging
import json
from superdesk.io.registry import register_feeding_service, register_feeding_service_parser
from superdesk.io.feeding_services.http_base_service import HTTPFeedingServiceBase
from superdesk.errors import IngestApiError, SuperdeskIngestError
import requests
import superdesk
from superdesk.io.feed_parsers import nitf
from lxml import etree
from superdesk.utc import utcnow
from datetime import timedelta, datetime

logger = logging.getLogger(__name__)


class APMediaFeedingService(HTTPFeedingServiceBase):
    """
    Feeding Service class which can retrieve articles from Associated Press Media API
    """

    NAME = "ap media api"

    label = "AP Media API"

    fields = [
        {
            "id": "api_url",
            "type": "text",
            "label": "AP Media API URL",
            "required": True,
            "default_value": "https://api.ap.org/media/v/content/feed",
        },
        {
            "id": "products_url",
            "type": "text",
            "label": "AP Media API Products URL",
            "required": True,
            "default_value": "https://api.ap.org/media/v/account/plans",
        },
        {
            "id": "apikey",
            "type": "text",
            "label": "API Key",
            "placeholder": "API key for access to the API",
            "required": True,
        },
        {
            "id": "productList",
            "type": "text",
            "label": "Product List",
            "placeholder": "Use coma separated product id" "s for multiple products, empty for all ",
            "required": False,
        },
        {"id": "availableProducts", "type": "text", "label": "All Available Products", "readonly": True},
        {"id": "next_link", "type": "text", "label": "Next Link", "readonly": True},
        {
            "id": "recoverytime",
            "type": "text",
            "label": "Number of hours to recover",
            "default_value": "",
            "placeholder": "Specifying a number will restart the feed from that number of hours in the past",
        },
    ]

    HTTP_AUTH = False

    HTTP_TIMEOUT = 40

    def config_test(self, provider=None):
        self._get_products(provider)
        original = superdesk.get_resource_service("ingest_providers").find_one(req=None, _id=provider.get("_id"))
        # If there has been a change in the required products then reset the next link
        if original and (
            original.get("config", {}).get("productList", "") != provider.get("config", {}).get("productList", "")
            or original.get("config", {}).get("recoverytime", "") != provider.get("config", {}).get("recoverytime", "")
        ):
            provider["config"]["next_link"] = None

    def _get_products(self, provider):
        """
        Get the products that are available for the API Key, effectively ensuring that the key is valid and provide an
         indication of the product codes available in the UI.
        :param provider:
        :return:
        """
        api_key = provider.get("config", {}).get("apikey")
        r = requests.get(
            provider.get("config", {}).get("products_url") + "?apikey={}".format(api_key),
            timeout=self.HTTP_TIMEOUT,
            verify=False,
            allow_redirects=True,
        )
        r.raise_for_status()
        productList = []
        products = json.loads(r.text)
        for plan in products.get("data", {}).get("plans"):
            for entitlement in plan.get("entitlements"):
                productList.append("{}".format(entitlement.get("id")))
        provider["config"]["availableProducts"] = ",".join(productList)

    def prepare_href(self, href, mimetype=None):
        href = (
            href + "&apikey=" + self.provider.get("config", {}).get("apikey")
            if "?" in href
            else href + "?apikey=" + self.provider.get("config", {}).get("apikey")
        )
        return href

    def _update(self, provider, update):
        self.HTTP_URL = provider.get("config", {}).get("api_url", "")
        self.provider = provider

        # Set the apikey parameter we're going to use it on all calls
        params = dict()
        params["apikey"] = provider.get("config", {}).get("apikey")

        # Use the next link if one is available in the config
        if provider.get("config", {}).get("next_link"):
            r = self.get_url(
                url=provider.get("config", {}).get("next_link"), params=params, verify=False, allow_redirects=True
            )
            r.raise_for_status()
        else:
            id_list = provider.get("config", {}).get("productList", "").strip()
            recovery_time = provider.get("config", {}).get("recoverytime", "1")
            recovery_time = recovery_time.strip() if recovery_time else ""
            if recovery_time == "":
                recovery_time = "1"
            start = datetime.strftime(utcnow() - timedelta(hours=int(recovery_time)), "%Y-%m-%dT%H:%M:%SZ")
            # If there has been a list of products defined then we format them for the request, if not all
            # allowed products will be returned.
            if id_list:
                # we remove spaces and empty values from id_list to do a clean list
                id_list = " OR ".join([id_.strip() for id_ in id_list.split(",") if id_.strip()])
                params["q"] = "productid:(" + id_list + ") AND mindate:>{}".format(start)
            else:
                params["q"] = "mindate:>{}".format(start)
            params["page_size"] = "100"
            params["versions"] = "all"

            logger.info("AP Media Start/Recovery time: {} params {}".format(recovery_time, params))
            r = self.get_url(params=params, verify=False, allow_redirects=True)
            r.raise_for_status()
        try:
            response = json.loads(r.text)
        except Exception:
            raise IngestApiError.apiRequestError(Exception("error parsing response"))

        nextLink = response.get("data", {}).get("next_page")
        # Got the same next link as last time so nothing new
        if nextLink == provider.get("config", {}).get("next_link"):
            logger.info("Nothing new from AP Media")
            return []

        parser = self.get_feed_parser(provider)
        parsed_items = []
        for item in response.get("data", {}).get("items", []):
            try:
                # Get the item meta data
                logger.info(
                    'Get AP meta data for "{}" uri: {}'.format(
                        item.get("item", {}).get("headline"), item.get("item", {}).get("uri")
                    )
                )
                r = self.api_get(item.get("item", {}).get("uri"), provider)
                complete_item = json.loads(r.text)

                # Get the nitf rendition of the item
                nitf_ref = (
                    complete_item.get("data", {}).get("item", {}).get("renditions", {}).get("nitf", {}).get("href")
                )
                if nitf_ref:
                    logger.info("Get AP nitf : {}".format(nitf_ref))
                    r = self.api_get(nitf_ref, provider)
                    root_elt = etree.fromstring(r.content)
                    nitf_item = nitf.NITFFeedParser().parse(root_elt)
                    complete_item["nitf"] = nitf_item
                else:
                    if item.get("item", {}).get("type") == "text":
                        logger.warning("No NITF for story {}".format(item.get("item", {}).get("uri")))

                associations = complete_item["data"]["item"].get("associations")
                if associations:
                    complete_item["associations"] = {}
                    for key, assoc in associations.items():
                        logger.info('Get AP association "%s"', assoc.get("headline"))
                        related_json = self.api_get(assoc["uri"], provider).json()
                        complete_item["associations"][key] = related_json

                parsed_items.append(parser.parse(complete_item, provider))

            # Any exception processing an indivisual item is swallowed
            except Exception as ex:
                logger.exception(ex)

        # Save the link for next time
        upd_provider = provider.get("config")
        upd_provider["next_link"] = nextLink
        upd_provider["recoverytime"] = None
        update["config"] = upd_provider

        return [parsed_items]

    def api_get(self, url, provider):
        resp = self.get_url(
            url=url, params={"apikey": provider["config"]["apikey"]}, verify=False, allow_redirects=True
        )
        resp.raise_for_status()
        return resp


register_feeding_service(APMediaFeedingService)



"""Geonames utils."""

import requests

from urllib.parse import urljoin
from urllib3.util.retry import Retry

from flask import current_app as app

session = requests.Session()
retries = Retry(total=3, backoff_factor=0.1)
adapter = requests.adapters.HTTPAdapter(max_retries=retries)
session.mount("http://", adapter)
session.mount("https://", adapter)


def get_geonames_state_qcode(place):
    return "iso3166-2:%s-%s" % (place.get("country_code"), place.get("state_code"))


def get_geonames_country_qcode(place):
    return "iso3166-1a2:%s" % place.get("country_code")


def get_geonames_qcode(place):
    if place.get("feature_class", "").upper() == "A":
        if place.get("state_code"):  # state
            return get_geonames_state_qcode(place)
        else:  # country
            return get_geonames_country_qcode(place)
    return "geonames:%s" % place.get("code", "0")


def format_geoname_item(item):
    geo = {
        "scheme": "geonames",
        "code": str(item["geonameId"]),
        "name": item["name"],
        "state": item.get("adminName1", ""),
        "region": item.get("adminName2", ""),
        "country": item.get("countryName", ""),
        "state_code": item.get("adminCode1", ""),
        "region_code": item.get("adminCode2", ""),
        "country_code": item.get("countryCode", ""),
        "continent_code": item.get("continentCode", ""),
        "feature_class": item.get("fcl"),
        "location": {
            "lat": float(item["lat"]),
            "lon": float(item["lng"]),
        },
    }

    if item.get("timezone"):
        geo["tz"] = item["timezone"].get("timeZoneId")

    return geo


def geonames_request(service, service_params):
    params = [
        ("type", "json"),
        ("username", app.config.get("GEONAMES_USERNAME", "")),
    ]

    if app.config.get("GEONAMES_TOKEN"):
        params.append(("token", app.config["GEONAMES_TOKEN"]))

    params.extend(service_params)
    url = urljoin(app.config["GEONAMES_URL"], service)
    res = session.get(url, params=params, timeout=10)
    if res.status_code != 200:
        res.raise_for_status()
    return res.json()



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license


import os
from unittest.mock import create_autospec
from superdesk.tests import TestCase
import ftplib
from apps.publish import init_app
from superdesk.publish.transmitters.ftp import FTPPublishService
import io
from unittest import mock


ASSOCIATIONS = {
    "featuremedia": {
        "renditions": {
            "16-9": {
                "media": "5e448ee9016d1f63a92f040b",
                "width": 1280,
                "href": "/5e448ee9016d1f63a92f040b.jpg",
                "height": 720,
                "mimetype": "image/jpeg",
                "poi": {"y": 1458, "x": 1331},
            },
            "4-3": {
                "media": "5e448ee8016d1f63a92f0408",
                "width": 800,
                "href": "/5e448ee8016d1f63a92f0408.jpg",
                "height": 600,
                "mimetype": "image/jpeg",
                "poi": {"y": 1502, "x": 1162},
            },
            "original": {
                "media": "5e448e47016d1f63a92f03b8",
                "width": 5568,
                "href": "/5e448e47016d1f63a92f03b8.jpg",
                "height": 3712,
                "mimetype": "image/jpeg",
                "poi": {"y": 1781, "x": 1893},
            },
        }
    },
    "editor_0": {
        "priority": 6,
        "subject": [{"code": "02000000", "name": "crime, law and justice"}],
        "body_text": "pic alt",
        "byline": "The Great Unwashed",
        "renditions": {
            "16-9": {
                "media": "5e448dd1016d1f63a92f039e",
                "width": 1280,
                "href": "/5e448dd1016d1f63a92f039e.png",
                "height": 720,
                "mimetype": "image/png",
            },
            "4-3": {
                "media": "5e448dd1016d1f63a92f0398",
                "width": 800,
                "href": "/5e448dd1016d1f63a92f0398.png",
                "height": 600,
                "mimetype": "image/png",
            },
            "original": {
                "media": "5e448dd1016d1f63a92f0393",
                "width": 1615,
                "href": "/upload-raw/5e448dd1016d1f63a92f0393.png",
                "height": 1026,
                "mimetype": "image/png",
            },
        },
    },
}


class NotFoundResponse:
    status_code = 404


class TestMedia(io.BytesIO):
    _id = "media-id"
    filename = "foo.txt"
    mimetype = "text/plain"


def mockGet(self, _id, resource=None):
    return b"binary"


class FTPPublishServiceTestCase(TestCase):
    def setUp(self):
        init_app(self.app)

    item = {"item_id": "abc", "format": "NITF", "formatted_item": "1234567890"}

    def is_item_loaded(self, url, uploaded_filename):
        config = FTPPublishService().config_from_url(url)
        with ftplib.FTP(config.get("host")) as ftp:
            ftp.login(config.get("username"), config.get("password"))
            ftp.cwd(config.get("path", ""))
            ftp.set_pasv(config.get("passive", False))

            for filename, facts in ftp.mlsd():
                if filename == uploaded_filename:
                    return True
            return False

    def test_it_can_connect(self):
        service = FTPPublishService()

        if "FTP_URL" not in os.environ:
            return

        config = service.config_from_url(os.environ["FTP_URL"])
        self.item["destination"] = {"config": config}

        self.assertEqual("test", config["path"])
        self.assertEqual("localhost", config["host"])

        service._transmit(self.item, subscriber={"config": config})
        self.assertTrue(self.is_item_loaded(config, "abc.ntf"))

    @mock.patch("ftplib.FTP", autospec=True)
    @mock.patch("superdesk.storage.ProxyMediaStorage.get", mockGet)
    @mock.patch("superdesk.publish.transmitters.ftp.get_renditions_spec", return_value={"16-9": {}, "4-3": {}})
    def test_with_associations(self, mock_ftp_constructor, *args):
        item = {
            "associations": {
                "featuremedia": ASSOCIATIONS["featuremedia"],
            }
        }

        service = FTPPublishService()

        service._copy_published_media_files(item, mock_ftp_constructor)

        mock_ftp_constructor.storbinary.assert_any_call("STOR 5e448e47016d1f63a92f03b8.jpg", b"binary")
        mock_ftp_constructor.storbinary.assert_any_call("STOR 5e448ee8016d1f63a92f0408.jpg", b"binary")
        mock_ftp_constructor.storbinary.assert_any_call("STOR 5e448ee9016d1f63a92f040b.jpg", b"binary")

    @mock.patch("superdesk.ftp.ftplib.FTP", autospec=True)
    @mock.patch("superdesk.storage.ProxyMediaStorage.get", mockGet)
    @mock.patch("superdesk.publish.transmitters.ftp.get_renditions_spec", return_value={"16-9": {}, "4-3": {}})
    def test_with_association_and_embed(self, mock_ftp_constructor, *args):
        item = {"associations": ASSOCIATIONS}

        service = FTPPublishService()
        service._copy_published_media_files(item, mock_ftp_constructor)

        mock_ftp_constructor.storbinary.assert_any_call("STOR 5e448e47016d1f63a92f03b8.jpg", b"binary")
        mock_ftp_constructor.storbinary.assert_any_call("STOR 5e448ee8016d1f63a92f0408.jpg", b"binary")
        mock_ftp_constructor.storbinary.assert_any_call("STOR 5e448ee9016d1f63a92f040b.jpg", b"binary")
        mock_ftp_constructor.storbinary.assert_any_call("STOR 5e448dd1016d1f63a92f0393.png", b"binary")
        mock_ftp_constructor.storbinary.assert_any_call("STOR 5e448dd1016d1f63a92f0398.png", b"binary")
        mock_ftp_constructor.storbinary.assert_any_call("STOR 5e448dd1016d1f63a92f039e.png", b"binary")

    @mock.patch("superdesk.publish.transmitters.ftp.ftp_connect")
    @mock.patch("superdesk.storage.ProxyMediaStorage.get", mockGet)
    def test_publish_non_ninjs_item_assoc(self, ftp_connect_mock, *args):
        service = FTPPublishService()
        queue_item = {
            "item_id": "someid",
            "item_version": 3,
            "formatted_item": "<?xml ...>",  # something json won't handle
            "destination": {"config": {"push_associated": True}},
        }

        ftp_mock = create_autospec(ftplib.FTP)()
        context_mock = mock.Mock()
        context_mock.__enter__ = mock.Mock(return_value=ftp_mock)
        context_mock.__exit__ = mock.Mock(return_value=None)
        ftp_connect_mock.return_value = context_mock

        self.app.data.insert(
            "published",
            [
                {
                    "item_id": "someid",
                    "_current_version": 3,
                    "associations": ASSOCIATIONS,
                }
            ],
        )

        subscriber = {}
        service._transmit(queue_item, subscriber)

        ftp_mock.storbinary.assert_any_call("STOR 5e448e47016d1f63a92f03b8.jpg", b"binary")
        ftp_mock.storbinary.assert_any_call("STOR 5e448dd1016d1f63a92f0393.png", b"binary")



import logging
from urllib.parse import urljoin

import requests
from bson import json_util
from flask import current_app as app

from superdesk.errors import SuperdeskApiError

logger = logging.getLogger(__name__)


class VideoEditorWrapper:
    """Video Service for integating with video server"""

    label = "Video Editor"

    def __init__(self):
        self.session = requests.Session()
        self.session.headers.update({"User-Agent": "superdesk"})

    def get_base_url(self):
        return app.config.get("VIDEO_SERVER_URL") + "/projects/"

    def create(self, file):
        return self._post(file)

    def edit(self, _id, updates):
        project = self._get(_id)
        response = {}
        if project.get("version", 1) == 1:
            response = self._duplicate(_id)
            _id = response.get("_id", _id)
        try:
            self._put(_id, updates)
        except SuperdeskApiError as ex:
            if response:
                self._delete(_id)
            raise ex
        return self._get(_id)

    def find_one(self, _id):
        return self._get(_id)

    def capture_preview_thumbnail(self, _id, position=0, crop=None, rotate=None):
        self._get_preview_thumbnail(_id, position, crop, rotate)
        return self._get(_id)

    def upload_preview_thumbnail(self, _id, file_storage):
        return self._post_preview_thumbnail(_id, file_storage)

    def create_timeline_thumbnails(self, _id, amount):
        return self._get_timeline_thumbnails(_id, amount)

    def _url(self, _id, resource=""):
        url = urljoin(self.get_base_url(), str(_id))
        if resource:
            return urljoin(url + "/", resource)
        return url

    def _get_response(self, resp, *status):
        json = json_util.loads(resp.text)
        if resp.status_code not in status:
            if resp.status_code == 400:
                raise SuperdeskApiError.badRequestError(message=json)
            if resp.status_code == 409:
                raise SuperdeskApiError.conflictError(message=json)
            if resp.status_code == 404:
                raise SuperdeskApiError.notFoundError(message=json)
            if resp.status_code == 500:
                raise SuperdeskApiError.internalError(message=json)
        return json

    def _get(self, project_id):
        try:
            resp = self.session.get(self._url(project_id))
            return self._get_response(resp, 200)
        except ConnectionError as ex:
            raise SuperdeskApiError(message=ex.args[0], status_code=500)

    def _get_paginate(self, page):
        try:
            resp = self.session.get(self.get_base_url(), params={"page": page})
            return self._get_response(resp, 200)
        except ConnectionError as ex:
            raise SuperdeskApiError(message=ex.args[0], status_code=500)

    def _post(self, file_storage):
        try:
            video_file = {"file": (file_storage.filename, file_storage.read(), file_storage.mimetype)}
            resp = self.session.post(self.get_base_url(), files=video_file)
            return self._get_response(resp, 201)
        except ConnectionError as ex:
            raise SuperdeskApiError(message=ex.args[0], status_code=500)

    def _duplicate(self, project_id):
        try:
            resp = self.session.post(self._url(project_id, "duplicate"))
            return self._get_response(resp, 201)
        except ConnectionError as ex:
            raise SuperdeskApiError(message=ex.args[0], status_code=500)

    def _put(self, project_id, updates):
        try:
            resp = self.session.put(self._url(project_id), json=updates)
            return self._get_response(resp, 202)
        except ConnectionError as ex:
            raise SuperdeskApiError(message=ex.args[0], status_code=500)

    def _delete(self, project_id):
        try:
            resp = self.session.delete(self._url(project_id))
            if resp.status_code != 204:
                logger.exception(resp.text)
                resp.raise_for_status()
            return True
        except ConnectionError as ex:
            raise SuperdeskApiError(message=ex.args[0], status_code=500)

    def _get_timeline_thumbnails(self, project_id, amount):
        try:
            params = {"type": "timeline", "amount": amount}
            resp = self.session.get(self._url(project_id, "thumbnails"), params=params)
            return self._get_response(resp, 202)
        except ConnectionError as ex:
            raise SuperdeskApiError(message=ex.args[0], status_code=500)

    def _post_preview_thumbnail(self, project_id, file_storage):
        try:
            video_file = {"file": (file_storage.filename, file_storage.read(), file_storage.mimetype)}
            resp = self.session.post(self._url(project_id, "thumbnails"), files=video_file)
            return self._get_response(resp, 200)
        except ConnectionError as ex:
            raise SuperdeskApiError(message=ex.args[0], status_code=500)

    def _get_preview_thumbnail(self, project_id, position=0, crop=None, rotate=None):
        try:
            params = {"type": "preview", "position": position}
            if crop:
                params["crop"] = crop
            if rotate:
                params["rotate"] = rotate
            resp = self.session.get(self._url(project_id, "thumbnails"), params=params)
            return self._get_response(resp, 202)
        except ConnectionError as ex:
            raise SuperdeskApiError(message=ex.args[0], status_code=500)



from lxml import etree


from .base import BasePackageElement


class Tags(BasePackageElement):
    """
    As Adobe's idml specification explains:
     ``
     The Tags.xml file contains the XML tag de nitions stored in the InDesign document,
     including unused tags.
     ``
    """

    DEFAULT_TAGS = {"Root": {"TagColor": "LightBlue"}}

    @property
    def filename(self):
        """
        Filename inside IDML package.
        Used as a filename for a file inside zip container.
        :return str: filename
        """
        return "XML/Tags.xml"

    def _build_etree(self):
        self._etree = etree.Element(etree.QName(self.XMLNS_IDPKG, "Tags"), nsmap={"idPkg": self.XMLNS_IDPKG})
        self._etree.set("DOMVersion", self.DOM_VERSION)
        self._add_xmltags()

    def _add_xmltags(self):
        """
        Add XMLTag(s)
        """
        tags = self.merge_attributes(self.DEFAULT_TAGS, self._attributes)

        for tag in tags:
            # XMLTag
            xmltag = etree.SubElement(
                self._etree,
                "XMLTag",
                attrib={
                    "Self": "XMLTag/{}".format(tag),
                    "Name": tag,
                },
            )
            # Properties
            properties = etree.SubElement(xmltag, "Properties")
            # TagColor
            tagcolor = etree.SubElement(properties, "TagColor", attrib={"type": "enumeration"})
            tagcolor.text = tags[tag]["TagColor"]



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license


import bson
import logging
from datetime import timedelta, timezone, datetime

from flask import current_app as app
from werkzeug.exceptions import HTTPException

import superdesk
from superdesk.activity import ACTIVITY_EVENT, notify_and_add_activity
from superdesk.celery_app import celery
from superdesk.celery_task_utils import get_lock_id
from superdesk.errors import ProviderError
from superdesk.io import get_feeding_service
from superdesk.io.registry import registered_feeding_services, registered_feed_parsers
from superdesk.io.iptc import subject_codes
from superdesk.lock import lock, unlock, touch
from superdesk.media.renditions import update_renditions, transfer_renditions
from superdesk.metadata.item import (
    GUID_NEWSML,
    GUID_FIELD,
    FAMILY_ID,
    ITEM_TYPE,
    CONTENT_TYPE,
    CONTENT_STATE,
    ITEM_STATE,
)
from superdesk.metadata.utils import generate_guid
from superdesk.notification import push_notification
from superdesk.utc import utcnow, get_expiry_date
from superdesk.workflow import set_default_state
from superdesk.errors import IngestFileError
from copy import deepcopy

UPDATE_SCHEDULE_DEFAULT = {"minutes": 5}
LAST_UPDATED = "last_updated"
LAST_INGESTED_ID = "last_ingested_id"
LAST_ITEM_UPDATE = "last_item_update"
IDLE_TIME_DEFAULT = {"hours": 0, "minutes": 0}

UPDATE_TTL = 1800

logger = logging.getLogger(__name__)


def is_service_and_parser_registered(provider):
    """Tests if the Feed Service and Feed Parser associated with are registered with application.

    :param provider:
    :type provider: dict :py:class:`superdesk.io.ingest_provider_model.IngestProviderResource`
    :return: True if both Feed Service and Feed Parser are registered. False otherwise.
    :rtype: bool
    """
    return (
        provider.get("feeding_service") in registered_feeding_services
        and provider.get("feed_parser") is None
        or provider.get("feed_parser") in registered_feed_parsers
    )


def is_scheduled(provider):
    """Test if given provider should be scheduled for update.

    :param provider: ingest provider
    """
    now = utcnow()
    last_updated = provider.get(LAST_UPDATED, now - timedelta(days=100))  # if never updated run now
    update_schedule = provider.get("update_schedule", UPDATE_SCHEDULE_DEFAULT)
    return last_updated + timedelta(**update_schedule) < now


def is_closed(provider):
    """Test if provider is closed.

    :param provider: ingest provider
    """
    return provider.get("is_closed", False)


def is_not_expired(item, delta):
    if item.get("expiry") or item.get("versioncreated"):
        try:
            expiry = item.get("expiry", item["versioncreated"] + delta)
        except OverflowError:  # this will never expire
            return True
        if expiry.tzinfo:
            return expiry > utcnow()
        else:
            return expiry > datetime.now()
    if not item.get("versioncreated"):  # can't say really
        return True
    return False


def filter_expired_items(provider, items):
    """Filter out expired items from the list of articles to be ingested.

    Filte both expired and `item['type'] not in provider['content_types']`.

    :param provider: Ingest Provider Details.
    :type provider: dict :py:class: `superdesk.io.ingest_provider_model.IngestProviderResource`
    :param items: list of items received from the provider
    :type items: list
    :return: list of items which can be saved into ingest collection
    :rtype: list
    """
    try:
        try:
            content_expiry = int(provider["content_expiry"])
        except ValueError:
            logger.warning("invalid content_expiry: content_expiry={value}".format(value=provider["content_expiry"]))
            del provider["content_expiry"]
            content_expiry = None
        except (TypeError, KeyError):
            content_expiry = None
        else:
            if content_expiry < 0:
                del provider["content_expiry"]
                content_expiry = None

        delta = timedelta(minutes=content_expiry or app.config["INGEST_EXPIRY_MINUTES"])
        filtered_items = [
            item
            for item in items
            if is_not_expired(item, delta) and item.get(ITEM_TYPE, "text") in provider.get("content_types", [])
        ]

        if len(items) != len(filtered_items):
            logger.warning(
                "Received {0} articles from provider {1}, but only {2} are eligible to be saved in ingest".format(
                    len(items), provider["name"], len(filtered_items)
                )
            )

        return filtered_items
    except Exception as ex:
        raise ProviderError.providerFilterExpiredContentError(ex, provider)


def get_provider_rule_set(provider):
    if provider.get("rule_set"):
        return superdesk.get_resource_service("rule_sets").find_one(_id=provider["rule_set"], req=None)


def get_provider_routing_scheme(provider):
    """Returns the ingests provider's routing scheme configuration.

    If provider has a routing scheme defined (i.e. scheme ID is not None), the
    scheme is fetched from the database. If not, nothing is returned.

    For all scheme rules that have a reference to a content filter defined,
    that filter's configuration is fetched from the database as well and
    embedded into the corresponding scheme rule.

    :param dict provider: ingest provider configuration

    :return: fetched provider's routing scheme configuration (if any)
    :rtype: dict or None
    """
    if not provider.get("routing_scheme"):
        return None

    schemes_service = superdesk.get_resource_service("routing_schemes")
    filters_service = superdesk.get_resource_service("content_filters")

    scheme = schemes_service.find_one(_id=provider["routing_scheme"], req=None)
    if not scheme:
        return None

    # for those routing rules that have a content filter defined,
    # get that filter from DB and embed it into the rule...
    rules_filters = ((rule, str(rule["filter"])) for rule in scheme["rules"] if rule.get("filter"))

    for rule, filter_id in rules_filters:
        content_filter = filters_service.find_one(_id=filter_id, req=None)
        rule["filter"] = content_filter

    return scheme


def get_task_ttl(provider):
    update_schedule = provider.get("update_schedule", UPDATE_SCHEDULE_DEFAULT)
    ttl = (
        update_schedule.get("minutes", 0) * 60
        + update_schedule.get("hours", 0) * 3600
        + update_schedule.get("seconds", 0)
    )
    if ttl < 15:
        ttl = 15
    return ttl


def get_is_idle(provider):
    last_item = provider.get(LAST_ITEM_UPDATE)
    idle_time = provider.get("idle_time", IDLE_TIME_DEFAULT)
    if isinstance(idle_time["hours"], datetime):
        idle_hours = 0
    else:
        idle_hours = idle_time["hours"]
    if isinstance(idle_time["minutes"], datetime):
        idle_minutes = 0
    else:
        idle_minutes = idle_time["minutes"]
    # there is an update time and the idle time is none zero
    if last_item and (idle_hours != 0 or idle_minutes != 0):
        if utcnow() > last_item + timedelta(hours=idle_hours, minutes=idle_minutes):
            return True
    return False


def get_task_id(provider):
    return "update-ingest-{0}-{1}".format(provider.get("name"), provider.get(superdesk.config.ID_FIELD))


def has_system_renditions(item):
    return all((rend in item["renditions"] for rend in ("viewImage", "baseImage", "thumbnail")))


def update_assoc_renditions(assoc, ingested):
    if ingested.get("renditions"):
        assoc.setdefault("renditions", {})
        for key, val in ingested["renditions"].items():
            assoc["renditions"][key] = val


class UpdateIngest(superdesk.Command):
    """Runs update for ingest providers.

    Example:
    ::

        $ python manage.py ingest:update
        $ python manage.py ingest:update --provider=aap-demo

    """

    option_list = (
        superdesk.Option("--provider", "-p", dest="provider_name"),
        superdesk.Option("--sync", "-s", dest="sync", action="store_true"),
    )

    def run(self, provider_name=None, sync=False):
        lookup = {} if not provider_name else {"name": provider_name}
        for provider in superdesk.get_resource_service("ingest_providers").get(req=None, lookup=lookup):
            if (
                not is_closed(provider)
                and is_service_and_parser_registered(provider)
                and (is_scheduled(provider) or sync)
            ):
                kwargs = {
                    "provider": provider,
                    "rule_set": get_provider_rule_set(provider),
                    "routing_scheme": get_provider_routing_scheme(provider),
                    "sync": sync,
                }

                if sync:
                    update_provider.apply(kwargs=kwargs)
                else:
                    update_provider.apply_async(expires=get_task_ttl(provider), kwargs=kwargs, serializer="eve/json")


def update_last_item_updated(update, items):
    if items:
        last_item_update = max(
            [item["versioncreated"] for item in items if item.get("versioncreated")], default=utcnow()
        )
        if not update.get(LAST_ITEM_UPDATE) or update[LAST_ITEM_UPDATE] < last_item_update:
            update[LAST_ITEM_UPDATE] = last_item_update


@celery.task(soft_time_limit=UPDATE_TTL)
def update_provider(provider, rule_set=None, routing_scheme=None, sync=False):
    """Fetch items from ingest provider, ingest them into Superdesk and update the provider.

    :param provider: Ingest Provider data
    :param rule_set: Translation Rule Set if one is associated with Ingest Provider.
    :param routing_scheme: Routing Scheme if one is associated with Ingest Provider.
    :param sync: Running in sync mode from cli.
    """
    lock_name = get_lock_id("ingest", provider["name"], provider[superdesk.config.ID_FIELD])

    if not lock(lock_name, expire=UPDATE_TTL + 10):
        if sync:
            logger.error("update is already running for %s", provider["name"])
        return

    try:
        feeding_service = get_feeding_service(provider["feeding_service"])
        update = {LAST_UPDATED: utcnow()}

        if sync:
            provider[LAST_UPDATED] = utcnow() - timedelta(days=9999)  # import everything again

        generator = feeding_service.update(provider, update)
        if isinstance(generator, list):
            generator = (items for items in generator)
        failed = None
        while True:
            try:
                if not touch(lock_name, expire=UPDATE_TTL):
                    logger.warning("lock expired while updating provider %s", provider[superdesk.config.ID_FIELD])
                    return
                items = generator.send(failed)
                failed = ingest_items(items, provider, feeding_service, rule_set, routing_scheme)
                update_last_item_updated(update, items)
            except StopIteration:
                break

        # Some Feeding Services update the collection and by this time the _etag might have been changed.
        # So it's necessary to fetch it once again. Otherwise, OriginalChangedError is raised.
        ingest_provider_service = superdesk.get_resource_service("ingest_providers")
        provider = ingest_provider_service.find_one(req=None, _id=provider[superdesk.config.ID_FIELD])
        ingest_provider_service.system_update(provider[superdesk.config.ID_FIELD], update, provider)

        if LAST_ITEM_UPDATE not in update and get_is_idle(provider):
            admins = superdesk.get_resource_service("users").get_users_by_user_type("administrator")
            notify_and_add_activity(
                ACTIVITY_EVENT,
                "Provider {{name}} has gone strangely quiet. Last activity was on {{last}}",
                resource="ingest_providers",
                user_list=admins,
                name=provider.get("name"),
                last=provider[LAST_ITEM_UPDATE].replace(tzinfo=timezone.utc).astimezone(tz=None).strftime("%c"),
            )

        logger.info("Provider {0} updated".format(provider[superdesk.config.ID_FIELD]))

        if LAST_ITEM_UPDATE in update:  # Only push a notification if there has been an update
            push_notification("ingest:update", provider_id=str(provider[superdesk.config.ID_FIELD]))
    except Exception as e:
        logger.error("Failed to ingest file: {error}".format(error=e))
        raise IngestFileError(3000, e, provider)
    finally:
        unlock(lock_name)


def process_anpa_category(item, provider):
    try:
        anpa_categories = superdesk.get_resource_service("vocabularies").find_one(req=None, _id="categories")
        if anpa_categories:
            for item_category in item["anpa_category"]:
                mapped_category = [
                    c
                    for c in anpa_categories["items"]
                    if c["is_active"] is True and item_category["qcode"].lower() == c["qcode"].lower()
                ]
                # if the category is not known to the system remove it from the item
                if len(mapped_category) == 0:
                    item["anpa_category"].remove(item_category)
                else:
                    item_category["name"] = mapped_category[0]["name"]
                    # make the case of the qcode match what we hold in our dictionary
                    item_category["qcode"] = mapped_category[0]["qcode"]
    except Exception as ex:
        raise ProviderError.anpaError(ex, provider)


def derive_category(item, provider):
    """Assuming that the item has at least one itpc subject use the vocabulary map to derive an anpa category.

    :param item:
    :return: An item with a category if possible
    """
    try:
        categories = []
        subject_map = superdesk.get_resource_service("vocabularies").find_one(req=None, _id="iptc_category_map")
        if subject_map:
            for entry in (map_entry for map_entry in subject_map["items"] if map_entry["is_active"]):
                for subject in item.get("subject", []):
                    if subject["qcode"] == entry["qcode"]:
                        if not any(c["qcode"] == entry["category"] for c in categories):
                            categories.append({"qcode": entry["category"]})
            if len(categories):
                item["anpa_category"] = categories
                process_anpa_category(item, provider)
    except Exception as ex:
        logger.exception(ex)


def process_iptc_codes(item, provider):
    """Ensures that the higher level IPTC codes are present by inserting them if missing.

    For example if given 15039001 (Formula One) make sure that 15039000 (motor racing) and 15000000 (sport)
    are there as well.

    :param item: A story item
    :return: A story item with possible expanded subjects
    """
    try:

        def iptc_already_exists(code):
            for entry in item["subject"]:
                if "qcode" in entry and code == entry["qcode"]:
                    return True
            return False

        for subject in item["subject"]:
            if "qcode" in subject and len(subject["qcode"]) == 8 and subject["qcode"].isdigit():
                top_qcode = subject["qcode"][:2] + "000000"
                if not iptc_already_exists(top_qcode):
                    try:
                        item["subject"].append({"qcode": top_qcode, "name": subject_codes[top_qcode]})
                    except KeyError:
                        logger.warning("missing qcode in subject_codes: {qcode}".format(qcode=top_qcode))
                        continue

                mid_qcode = subject["qcode"][:5] + "000"
                if not iptc_already_exists(mid_qcode):
                    try:
                        item["subject"].append({"qcode": mid_qcode, "name": subject_codes[mid_qcode]})
                    except KeyError:
                        logger.warning("missing qcode in subject_codes: {qcode}".format(qcode=mid_qcode))
                        continue
    except Exception as ex:
        raise ProviderError.iptcError(ex, provider)


def derive_subject(item):
    """Try to derive a subject using the anpa category vocabulary.

    :param item:
    :return:
    """
    try:
        category_map = superdesk.get_resource_service("vocabularies").find_one(req=None, _id="categories")
        if category_map:
            for cat in item["anpa_category"]:
                map_entry = next(
                    (code for code in category_map["items"] if code["qcode"] == cat["qcode"] and code["is_active"]),
                    None,
                )
                if map_entry and "subject" in map_entry:
                    item["subject"] = [
                        {"qcode": map_entry.get("subject"), "name": subject_codes[map_entry.get("subject")]}
                    ]
    except Exception as ex:
        logger.exception(ex)


def apply_rule_set(item, provider, rule_set=None):
    """Applies rules set on the item to be ingested into the system.

    If there's no rule set then the item will
    be returned without any change.

    :param item: Item to be ingested
    :param provider: provider object from whom the item was received
    :return: item
    """
    try:
        if rule_set is None and provider.get("rule_set") is not None:
            rule_set = superdesk.get_resource_service("rule_sets").find_one(_id=provider["rule_set"], req=None)

        if rule_set and "body_html" in item:
            body = item["body_html"]

            for rule in rule_set["rules"]:
                body = body.replace(rule["old"], rule["new"])

            item["body_html"] = body

        return item
    except Exception as ex:
        raise ProviderError.ruleError(ex, provider)


def ingest_cancel(item, feeding_service):
    """Given an item that has a pubstatus of canceled finds all versions of this item and mark them as canceled as well.

    Uses the URI to identify those items in ingest that are related to this cancellation.

    :param item:
    :return:
    """
    ingest_service = superdesk.get_resource_service(
        feeding_service.service if hasattr(feeding_service, "service") else "ingest"
    )
    lookup = {"uri": item.get("uri")}
    family_members = ingest_service.get_from_mongo(req=None, lookup=lookup)
    for relative in family_members:
        update = {"pubstatus": "canceled", ITEM_STATE: CONTENT_STATE.KILLED}
        ingest_service.patch(relative["_id"], update)


def ingest_items(items, provider, feeding_service, rule_set=None, routing_scheme=None):
    all_items = filter_expired_items(provider, items)
    items_dict = {doc[GUID_FIELD]: doc for doc in all_items}
    items_in_package = []
    failed_items = set()
    created_ids = []
    for item in [doc for doc in all_items if doc.get(ITEM_TYPE) == CONTENT_TYPE.COMPOSITE]:
        items_in_package = [
            ref["residRef"] for group in item.get("groups", []) for ref in group.get("refs", []) if "residRef" in ref
        ]

    for item in [doc for doc in all_items if doc.get(ITEM_TYPE) != CONTENT_TYPE.COMPOSITE]:
        ingested, ids = ingest_item(
            item,
            provider,
            feeding_service,
            rule_set,
            routing_scheme=routing_scheme if not item[GUID_FIELD] in items_in_package else None,
        )
        if ingested:
            created_ids = created_ids + ids
        else:
            failed_items.add(item[GUID_FIELD])
    for item in [doc for doc in all_items if doc.get(ITEM_TYPE) == CONTENT_TYPE.COMPOSITE]:
        for ref in [ref for group in item.get("groups", []) for ref in group.get("refs", []) if "residRef" in ref]:
            if ref["residRef"] in failed_items:
                failed_items.add(item[GUID_FIELD])
                continue

            ref.setdefault("location", "ingest")
            itemRendition = items_dict.get(ref["residRef"], {}).get("renditions")
            if itemRendition:
                ref.setdefault("renditions", itemRendition)
            ref[GUID_FIELD] = ref["residRef"]
            if items_dict.get(ref["residRef"]):
                ref["residRef"] = items_dict.get(ref["residRef"], {}).get(superdesk.config.ID_FIELD)
        if item[GUID_FIELD] in failed_items:
            continue
        ingested, ids = ingest_item(item, provider, feeding_service, rule_set, routing_scheme)
        if ingested:
            created_ids = created_ids + ids
        else:
            failed_items.add(item[GUID_FIELD])
    # sync mongo with ingest after all changes
    ingest_collection = feeding_service.service if hasattr(feeding_service, "service") else "ingest"
    ingest_service = superdesk.get_resource_service(ingest_collection)
    updated_items = ingest_service.find({"_id": {"$in": created_ids}}, max_results=len(created_ids))
    app.data._search_backend(ingest_collection).bulk_insert(ingest_collection, list(updated_items))
    if failed_items:
        logger.error("Failed to ingest the following items: %s", failed_items)
    return failed_items


def ingest_item(item, provider, feeding_service, rule_set=None, routing_scheme=None):
    items_ids = []
    try:
        ingest_collection = feeding_service.service if hasattr(feeding_service, "service") else "ingest"
        ingest_service = superdesk.get_resource_service(ingest_collection)

        # determine if we already have this item
        old_item = ingest_service.find_one(guid=item[GUID_FIELD], req=None)

        if not old_item:
            item.setdefault(superdesk.config.ID_FIELD, generate_guid(type=GUID_NEWSML))
            item[FAMILY_ID] = item[superdesk.config.ID_FIELD]

        item["ingest_provider"] = str(provider[superdesk.config.ID_FIELD])
        item.setdefault("source", provider.get("source", ""))
        item.setdefault("uri", item[GUID_FIELD])  # keep it as original guid

        if item.get("profile"):
            try:
                item["profile"] = bson.ObjectId(item["profile"])
            except bson.errors.InvalidId:
                pass

        set_default_state(item, CONTENT_STATE.INGESTED)
        item["expiry"] = get_expiry_date(
            provider.get("content_expiry") or app.config["INGEST_EXPIRY_MINUTES"], item.get("versioncreated")
        )

        if "anpa_category" in item:
            process_anpa_category(item, provider)

        if "subject" in item:
            if not app.config.get("INGEST_SKIP_IPTC_CODES", False):
                # FIXME: temporary fix for SDNTB-344, need to be removed once SDESK-439 is implemented
                process_iptc_codes(item, provider)
            if "anpa_category" not in item:
                derive_category(item, provider)
        elif "anpa_category" in item:
            derive_subject(item)

        apply_rule_set(item, provider, rule_set)

        if item.get("pubstatus", "") == "canceled":
            item[ITEM_STATE] = CONTENT_STATE.KILLED
            ingest_cancel(item, feeding_service)

        rend = item.get("renditions", {})
        if rend:
            baseImageRend = rend.get("baseImage") or next(iter(rend.values()))
            if baseImageRend and not baseImageRend.get("media"):  # if there is media should be processed already
                href = feeding_service.prepare_href(baseImageRend["href"], rend.get("mimetype"))
                update_renditions(item, href, old_item)

        # if the item has associated media
        for key, assoc in item.get("associations", {}).items():
            set_default_state(assoc, CONTENT_STATE.INGESTED)
            # wire up the id of the associated feature media to the ingested one
            guid = assoc.get("guid")
            assoc_name = assoc.get("headline") or assoc.get("slugline") or guid
            if guid:
                ingested = ingest_service.find_one(req=None, guid=guid)
                logger.info("assoc ingested before %s", assoc_name)
                if ingested is not None:
                    assoc["_id"] = ingested["_id"]
                    if is_new_version(assoc, ingested) and assoc.get("renditions"):  # new version
                        logger.info("new assoc version - re-transfer renditions for %s", assoc_name)
                        transfer_renditions(assoc["renditions"])
                    else:
                        logger.info("same/old version - use already fetched renditions for %s", assoc_name)
                        update_assoc_renditions(assoc, ingested)
                else:  # there is no such item in the system - ingest it
                    if assoc.get("renditions") and has_system_renditions(assoc):  # all set, just download
                        logger.info("new association  with system renditions - transfer %s", assoc_name)
                        transfer_renditions(assoc["renditions"])
                    status, ids = ingest_item(assoc, provider, feeding_service, rule_set)
                    if status:
                        assoc["_id"] = ids[0]
                        items_ids.extend(ids)
                        ingested = ingest_service.find_one(req=None, _id=ids[0])
                        update_assoc_renditions(assoc, ingested)
            elif assoc.get("residRef"):
                item["associations"][key] = resolve_ref(assoc)

        new_version = True
        if old_item:
            new_version = is_new_version(item, old_item)
            updates = deepcopy(item)
            ingest_service.patch_in_mongo(old_item[superdesk.config.ID_FIELD], updates, old_item)
            item.update(old_item)
            item.update(updates)
            items_ids.append(item["_id"])
        else:
            if item.get("ingest_provider_sequence") is None:
                ingest_service.set_ingest_provider_sequence(item, provider)
            try:
                items_ids.extend(ingest_service.post_in_mongo([item]))
            except HTTPException as e:
                logger.error("Exception while persisting item in %s collection: %s", ingest_collection, e)
                raise e

        if routing_scheme and new_version:
            routed = ingest_service.find_one(_id=item[superdesk.config.ID_FIELD], req=None)
            superdesk.get_resource_service("routing_schemes").apply_routing_scheme(routed, provider, routing_scheme)

    except Exception as ex:
        logger.exception(ex)
        ProviderError.ingestItemError(ex, provider, item=item)
        return False, []
    return True, items_ids


def resolve_ref(assoc):
    """Resolve reference to existing item."""
    uri = assoc.pop("residRef")
    item = superdesk.get_resource_service("archive").find_one(req=None, uri=uri)
    return item


NEW_VERSION_IGNORE_FIELS = ("expiry",)


def is_new_version(item, old_item):
    # explicit version info
    for field in ("version", "versioncreated"):
        if item.get(field) and old_item.get(field):
            try:
                return int(item[field], 10) > int(old_item[field], 10)
            except (ValueError, TypeError):
                try:
                    return item[field] > old_item[field]
                except TypeError as ex:
                    logger.exception(ex)
                    # can't compare the values, assuming these are different
                    return True
    # no version info, check content
    for field in item:
        if field in NEW_VERSION_IGNORE_FIELS or item[field] is None:
            continue
        if not old_item.get(field) or item[field] != old_item[field]:
            return True
    return False


superdesk.command("ingest:update", UpdateIngest())



import pytz
import warnings

from datetime import datetime
from unittest.mock import patch
from superdesk.tests import TestCase
from superdesk.io.feeding_services import FeedingService
from superdesk.errors import SuperdeskIngestError


class TestFeedingService(FeedingService):
    Name = "test_feeding"

    def _update(self, provider, update):
        pass


class FeedingServiceParserTestCase(TestCase):
    def test_restricted_parser(self):
        feeding_service = TestFeedingService()
        with patch.dict(
            "superdesk.io.feeding_services.restricted_feeding_service_parsers", {"test_feeding": {"ninjs": True}}
        ):
            with self.assertRaises(SuperdeskIngestError):
                feeding_service.config_test({"feed_parser": "foo", "feeding_service": "test_feeding"})

    def test_add_timestamps_warning(self):
        item = {}
        feeding_service = TestFeedingService()
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            feeding_service.add_timestamps(item)
        assert len(w) == 1
        assert issubclass(w[0].category, DeprecationWarning)

    def test_localize_timestamps(self):
        item = {"firstcreated": datetime(2019, 10, 15, 10, 59, 0)}
        self.assertIsNone(item["firstcreated"].tzinfo)
        feeding_service = TestFeedingService()
        feeding_service.localize_timestamps(item)
        self.assertEqual(pytz.utc, item["firstcreated"].tzinfo)



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license


import os

import superdesk
from superdesk.io.feed_parsers.rfc822 import EMailRFC822FeedParser
from superdesk.tests import TestCase, setup
from superdesk.errors import IngestEmailError
from superdesk.users.services import UsersService


class RFC822TestCase(TestCase):
    filename = "simple_email.txt"

    def setUp(self):
        setup(context=self)
        with self.app.app_context():
            # mock one user:
            user_service = UsersService("users", backend=superdesk.get_backend())
            self.user_id = user_service.create(
                [{"name": "user", "user_type": "administrator", "email": "asender@a.com.au"}]
            )[0]

            provider = {"name": "Test"}
            dirname = os.path.dirname(os.path.realpath(__file__))
            fixture = os.path.normpath(os.path.join(dirname, "../fixtures", self.filename))
            with open(fixture, mode="rb") as f:
                bytes = f.read()
            parser = EMailRFC822FeedParser()
            self.items = parser.parse([(1, bytes)], provider)

    def test_headline(self):
        self.assertEqual(self.items[0]["headline"], "Test message 1234")

    def test_body(self):
        self.assertEqual(
            self.items[0]["body_html"].strip(), '<div dir="ltr">body text<br clear="all"/><div>&#13;\n</div></div>'
        )

    def test_from(self):
        self.assertEqual(self.items[0]["original_source"], "a sender <asender@a.com.au>")
        self.assertEqual(self.items[0]["original_creator"], self.user_id)


class RFC822ComplexTestCase(TestCase):
    filename = "composite_email.txt"

    def setUp(self):
        setup(context=self)
        with self.app.app_context():
            provider = {"name": "Test"}
            dirname = os.path.dirname(os.path.realpath(__file__))
            fixture = os.path.normpath(os.path.join(dirname, "../fixtures", self.filename))
            with open(fixture, mode="rb") as f:
                bytes = f.read()
            parser = EMailRFC822FeedParser()
            self.items = parser.parse([(1, bytes)], provider)

    def test_composite(self):
        self.assertEqual(len(self.items), 3)
        for item in self.items:
            self.assertIn("versioncreated", item)

    def test_from(self):
        self.assertEqual(self.items[0]["original_source"], "someone <a@a.com.au>")
        self.assertNotIn("original_creator", self.items[0])


class RFC822OddCharSetTestCase(TestCase):
    filename = "odd_charset_email.txt"

    def setUp(self):
        setup(context=self)
        with self.app.app_context():
            provider = {"name": "Test"}
            dirname = os.path.dirname(os.path.realpath(__file__))
            fixture = os.path.normpath(os.path.join(dirname, "../fixtures", self.filename))
            with open(fixture, mode="rb") as f:
                bytes = f.read()
            parser = EMailRFC822FeedParser()
            self.items = parser.parse([(1, bytes)], provider)

    def test_headline(self):
        # This tests a subject that fails to decode but we just try a string conversion
        self.assertEqual(self.items[0]["headline"], "=?windows-1252?Q?TravTalk���s_Special_for_TAAI_convention?=")

    def test_body(self):
        self.assertRegex(self.items[0]["body_html"], '<span lang="EN-AU">')


class RFC822CharSetInSubjectTestCase(TestCase):
    filename = "charset_in_subject_email.txt"

    def setUp(self):
        setup(context=self)
        with self.app.app_context():
            provider = {"name": "Test"}
            dirname = os.path.dirname(os.path.realpath(__file__))
            fixture = os.path.normpath(os.path.join(dirname, "../fixtures", self.filename))
            with open(fixture, mode="rb") as f:
                bytes = f.read()
            parser = EMailRFC822FeedParser()
            self.items = parser.parse([(1, bytes)], provider)

    def test_headline(self):
        # This test a subject that has a charset that decodes correctly
        self.assertEqual(self.items[0]["headline"], "Google Apps News March 2015")


class RFC822FormattedEmailTestCase(TestCase):
    def setUp(self):
        setup(context=self)
        with self.app.app_context():
            self.app.data.insert(
                "users",
                [
                    {
                        "_id": 123,
                        "name": "user",
                        "user_type": "administrator",
                        "email": "eharvey@aap.com.au",
                        "byline": "E Harvey",
                    }
                ],
            )
            self.app.data.insert("desks", [{"_id": 1, "name": "new zealand"}])
            self.app.data.insert(
                "vocabularies",
                [
                    {"_id": "locators", "items": [{"is_active": True, "name": "ADV", "qcode": "ADV", "group": ""}]},
                    {"_id": "priority", "items": [{"is_active": True, "name": "Urgent", "qcode": "3"}]},
                ],
            )

            self.provider = {"name": "Test", "config": {"formatted": True}}

    def test_parsed_values(self):
        filename = "googleform.txt"
        dirname = os.path.dirname(os.path.realpath(__file__))
        fixture = os.path.normpath(os.path.join(dirname, "../fixtures", filename))
        with open(fixture, mode="rb") as f:
            bytes = f.read()
        parser = EMailRFC822FeedParser()

        self.items = parser.parse([(1, bytes)], self.provider)
        self.assertEqual(self.items[0]["headline"], "TEST NZ HEADER")
        self.assertEqual(self.items[0]["task"]["desk"], 1)
        self.assertEqual(self.items[0]["original_creator"], 123)
        self.assertEqual(self.items[0]["urgency"], 1)
        self.assertEqual(self.items[0]["dateline"]["text"], "AUCKLAND, May 5 AAP -")
        self.assertEqual(self.items[0]["place"][0]["qcode"], "ADV")
        self.assertEqual(self.items[0]["flags"]["marked_for_legal"], True)
        self.assertEqual(self.items[0]["priority"], 3)
        self.assertEqual(self.items[0]["byline"], "E Harvey")

    def test_parsed_values_Email_Address(self):
        filename = "googleform1.txt"
        dirname = os.path.dirname(os.path.realpath(__file__))
        fixture = os.path.normpath(os.path.join(dirname, "../fixtures", filename))
        with open(fixture, mode="rb") as f:
            bytes = f.read()
        parser = EMailRFC822FeedParser()
        self.items = parser.parse([(1, bytes)], self.provider)
        self.assertEqual(self.items[0]["headline"], "Arnold 'worried' about maroon-clad Roar")
        self.assertEqual(self.items[0]["task"]["desk"], 1)
        self.assertEqual(self.items[0]["original_creator"], 123)
        self.assertEqual(self.items[0]["urgency"], 2)
        self.assertEqual(self.items[0]["dateline"]["text"], "BRISBANE, Nov 18 AAP -")
        self.assertEqual(self.items[0]["byline"], "E Harvey")


class RFC822JsonEmailTestCase(TestCase):
    vocab = [{"_id": "categories", "items": [{"is_active": True, "name": "Domestic Sport", "qcode": "s"}]}]
    desk = [{"_id": 1, "name": "Brisbane"}]
    user = [{"_id": 1, "email": "mock@mail.com.au", "byline": "A Mock Up", "sign_off": "TA"}]

    def setUp(self):
        self.app.data.insert("vocabularies", self.vocab)
        self.app.data.insert("desks", self.desk)
        self.app.data.insert("users", self.user)
        with self.app.app_context():
            self.provider = {"name": "Test", "config": {"formatted": True}}

    def test_formatted_email_parser(self):
        filename = "json-email.txt"
        dirname = os.path.dirname(os.path.realpath(__file__))
        fixture = os.path.join(dirname, "../fixtures", filename)
        with open(fixture, mode="rb") as f:
            bytes = f.read()
        parser = EMailRFC822FeedParser()
        self.items = parser.parse([(1, bytes)], self.provider)

        self.assertEqual(self.items[0]["priority"], 5)
        self.assertEqual(self.items[0]["sign_off"], "TA")
        self.assertEqual(self.items[0]["anpa_category"], [{"qcode": "s"}])
        self.assertEqual(self.items[0]["body_html"], "<p>Lorem ipsum</p>")
        self.assertEqual(self.items[0]["abstract"], "Abstract-2")
        self.assertEqual(self.items[0]["headline"], "Headline-2")
        self.assertEqual(self.items[0]["original_creator"], 1)
        self.assertEqual(self.items[0]["task"]["desk"], 1)
        self.assertEqual(self.items[0]["urgency"], 4)
        self.assertEqual(self.items[0]["type"], "text")
        self.assertEqual(self.items[0]["guid"], "<001a1140cd6ecd9de8051fab814d@google.com>")
        self.assertEqual(self.items[0]["original_source"], "mock@mail.com.au")
        self.assertEqual(self.items[0]["slugline"], "Slugline-2")
        self.assertEqual(self.items[0]["byline"], "A Mock Up")

    def test_bad_user(self):
        filename = "json-email-bad-user.txt"
        dirname = os.path.dirname(os.path.realpath(__file__))
        fixture = os.path.join(dirname, "../fixtures", filename)
        with open(fixture, mode="rb") as f:
            bytes = f.read()
        parser = EMailRFC822FeedParser()

        try:
            with self.assertRaises(IngestEmailError) as exc_context:
                self.items = parser.parse([(1, bytes)], self.provider)
        except Exception:
            self.fail("Expected exception type was not raised.")

        ex = exc_context.exception
        self.assertEqual(ex.code, 6001)



import superdesk

from flask import current_app as app
from apps.auth import is_current_user_admin


class ConfigResource(superdesk.Resource):
    schema = {
        "_id": {"type": "string", "required": True},
        "val": {"type": "dict", "schema": {}, "allow_unknown": True},
    }

    item_url = 'regex("[\w.:_-]+")'
    item_methods = ["GET"]
    resource_methods = ["POST"]


class ConfigService(superdesk.Service):
    def find_one(self, req, **lookup):
        item = super().find_one(req, **lookup)
        if item:
            return item
        else:
            return {"_id": lookup.get("_id"), "val": None}

    def create(self, docs, **kwargs):
        ids = []
        for doc in docs:
            self.set(doc["_id"], doc.get("val"))
            ids.append(doc["_id"])
        return ids

    def set(self, key, val, namespace="superdesk"):
        coll = app.data.mongo.get_collection_with_write_concern("config", "config")
        coll.update_one({"_id": key}, {"$set": {"_id": key, "val": val}}, upsert=True)

    def get(self, key, namespace="superdesk"):
        return self.find_one(req=None, _id=key).get("val")

    def is_authorizes(self, user):
        return is_current_user_admin()


def init_app(app):
    superdesk.register_resource("config", ConfigResource, ConfigService, _app=app)
    superdesk.intrinsic_privilege("config", method=["POST"])



import re
import bson

from flask import current_app as app

from .user_mentions import notify_mentioned_users


# client is using @[display name](type:id)
USER_MENTIONS_REGEX = re.compile(r"@\[([^]]+)\]\(user:([a-f0-9]{24})\)")


def handle_inline_mentions(sender, updates, original):
    """Listen to item_update signal and send notifications to new inline user mentions."""
    updated = original.copy()
    updated.update(updates)
    comments = _get_inline_comments(updates)
    for comment in comments:
        if not comment.get("notified"):
            users = _get_mentioned_users(comment)
            if users:
                notify_mentioned_users(
                    [
                        {
                            "_id": "",
                            "item": original.get("_id"),
                            "text": _format_comment_text(comment),
                            "mentioned_users": {user: bson.ObjectId(user) for user in users},
                        }
                    ],
                    app.config.get("CLIENT_URL", "").rstrip("/"),
                    item=updated,
                )
            comment["notified"] = True


def _get_mentioned_users(comment):
    return [group[1] for group in USER_MENTIONS_REGEX.findall(comment.get("msg", ""))]


def _format_comment_text(comment):
    def repl(match):
        return match.group(1)

    return USER_MENTIONS_REGEX.sub(repl, comment.get("msg", ""))


def _get_inline_comments(updates):
    try:
        comments = []
        data = updates["fields_meta"]["body_html"]["draftjsState"][0]["blocks"][0]["data"].get(
            "__PUBLIC_API__comments", []
        )
        for val in data:
            comments.append(val)
            for reply in val.get("replies", []):
                comments.append(reply)
        return comments
    except (KeyError, IndexError):
        return []



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

import unittest
from .volume_cubic_yard_to_metric import cubic_yard_to_metric


class VolumeTestCase(unittest.TestCase):
    def test(self):
        text = (
            "Total volume is 100.50 cubic yards for this land. "
            "Total volume is 15.7 cubic yds for this land. "
            "Total volume is 1 Cubic Yard for this land. "
            "Total volume is 1-16 cu-yd for this land. "
            "Total volume is 1-16 cb. yds for this land. "
            "Total volume is 16.7-Cubic-yd for this land. "
            "Total volume is 16,500-cu. yd. for this land. "
        )
        item = {"body_html": text}
        res, diff = cubic_yard_to_metric(item)
        self.assertEqual(diff["100.50 cubic yards"], "100.50 cubic yards (76.84 cubic meter)")
        self.assertEqual(diff["15.7 cubic yds"], "15.7 cubic yds (12.0 cubic meter)")
        self.assertEqual(diff["1 Cubic Yard"], "1 Cubic Yard (0.76 cubic meter)")
        self.assertEqual(diff["1-16 cu-yd"], "1-16 cu-yd (0.76-12 cubic meter)")
        self.assertEqual(diff["1-16 cb. yds"], "1-16 cb. yds (0.76-12 cubic meter)")
        self.assertEqual(diff["16.7-Cubic-yd"], "16.7-Cubic-yd (12.8 cubic meter)")
        self.assertEqual(diff["16,500-cu. yd"], "16,500-cu. yd (12,615 cubic meter)")
        self.assertEqual(res["body_html"], item["body_html"])



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

import unittest
from .length_nautical_miles_to_metric import nautical_miles_to_metric


class MilesTestCase(unittest.TestCase):
    def test_m_to_km(self):
        text = (
            "Tom rides 125 nautical miles everyday. "
            "Tom rides 1 Nautical mile everyday. "
            "Tom rides 125 Nautical Miles everyday. "
            "Tom rides 125-nautical Miles everyday. "
            "Tom rides 12.5 nmi everyday. "
            "Tom rides 12.5nmi everyday. "
            "Tom rides 12.5-nmi everyday. "
            "Tom rides 12,500 nautical miles everyday. "
            "Tom rides 100-12500 Nautical Miles everyday. "
        )
        item = {"body_html": text}
        res, diff = nautical_miles_to_metric(item)
        self.assertEqual(diff["125 nautical miles"], "125 nautical miles (231.5 km)")
        self.assertEqual(diff["1 Nautical mile"], "1 Nautical mile (1.9 km)")
        self.assertEqual(diff["125 Nautical Miles"], "125 Nautical Miles (231.5 km)")
        self.assertEqual(diff["125-nautical Miles"], "125-nautical Miles (231.5 km)")
        self.assertEqual(diff["12.5 nmi"], "12.5 nmi (23.2 km)")
        self.assertEqual(diff["12.5nmi"], "12.5nmi (23.2 km)")
        self.assertEqual(diff["12.5-nmi"], "12.5-nmi (23.2 km)")
        self.assertEqual(diff["12,500 nautical miles"], "12,500 nautical miles (23,150 km)")
        self.assertEqual(diff["100-12500 Nautical Miles"], "100-12500 Nautical Miles (185.2-23,150 km)")
        self.assertEqual(res["body_html"], item["body_html"])
